


package gitzel.util;

import java.io.*;
import java.util.*;
import java.io.IOException;

/**
 * This class can read small scripts from various streams, readers etc. There
 * is no real parsing involved, rather the scripts consist of lines like this:
 * <p><pre>        command arg1 arg2 arg3 ...</pre>
 * <p>The getEntry(int line) method will return a string array that has the
 * command name in [0], and the args in the consecutive fields. getEntry(string)
 * should not be used directly but can be - getEntry("1-3") for example gets
 * you the 3rd argument of line 1.
 * Note that a program starts with line 0.
 * There are some commands that are hard-wired into this class:
 * <ul>
 *   <li>loop x - repeat next line x times (inf = endless loop)
 *   <li>goto x - jumps to line x
 * </ul>
 * <p>These commands are automatically executed when the iterator is used.
 * @author Ralf Gitzel
 * @version 1.0
 */
public class BasicScriptReader extends IniReader
{

/** The highest line number parsed yet. */
private int LineNum;

/** Reads the script from a buffered reader */
public BasicScriptReader(BufferedReader r) throws IOException
  {
  super(r);
  }

/** Reads the script from the InputStream. */
public BasicScriptReader(InputStream in) throws IOException
  {
  super(in);
  }

/** reads the script from the given string */
public BasicScriptReader(String data) throws IOException
  {
  super(data);
  }


/** Returns the number of lines this script has. */
public int getLineNumber()
  {
  return LineNum;
  }

/** returns a line slashed into command, and arguments */
public String[] getEntry(int line)
  {
  List temp = new ArrayList();
  String arg = null;
  int argnum = 0;
  while(true)
    {
    arg = getEntry(""+line+"-"+argnum);
    if (arg == null) break;
    temp.add(arg);
    argnum++;
    }
  Iterator it = temp.iterator();
  String ret[] = new String[temp.size()];
  argnum = 0;
  while (it.hasNext())
    {
    ret[argnum] = (String) it.next();
    argnum++;
    }
  return ret;
  }


/** The parsing of a line. */
public void parse(String line)
  {
  if (line == null) return;         // ignore eof-lines
  if (line.equals("")) return;      // ignore empty lines
  if (line.startsWith("#")) return; // ignore comment lines
  /////////////////////
  String end = line.trim(); // remove leading/trailing whitespace
  int argnum = 0;
  StringTokenizer parseBoy = new StringTokenizer(end);
  while (parseBoy.hasMoreTokens())
    {
    setEntry(""+LineNum+"-"+argnum, parseBoy.nextToken());
    argnum++;
    }
  LineNum++;
  }

/** Returns an iterator over this object. */
public Iterator iterator()
  {
  return new BSRIterator(this);
  }

/** The iterator class for this class. loop and goto are automatically executed
 *  when next is called. */
class BSRIterator implements Iterator
  {
  BasicScriptReader Base;

  /** The line that will be returned when next is called */
  int CurrLine = 0;

  /** used for loops. -1 means infinite loops */
  int repeat = 0;

  BSRIterator(BasicScriptReader base)
    {
    Base = base;
    }

  public boolean hasNext()
    {
    if (CurrLine < Base.getLineNumber()) return true;
    return false;
    }

  public Object next()
    {
    String[] data = Base.getEntry(CurrLine);
    if (data[0].equals("loop"))
      {
      // init loop
      if (data[1].equals("inf")) repeat = -1;
      else repeat = new Integer(data[1]).intValue();
      // move to the line to be looped
      CurrLine++;
      data = Base.getEntry(CurrLine); // already return next line
      }

    if (data[0].equals("goto"))
      {
      CurrLine = new Integer(data[1]).intValue();
      data = Base.getEntry(CurrLine); // already return next line
      }

    // if no loop goto next line
    if (repeat == 0) CurrLine++;
    else
      {
      // loops will count down - unless they are inf
      if (repeat != -1) repeat -= 1;
      }
    // no loop or goto...
    return data;
    }

  public void remove()
    {
    throw new UnsupportedOperationException("Remove not allowed!");
    }
  }

}

