
package gitzel.util;

import java.util.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;

/** This class can be used to customize tables. It provides methods for adding and
 *  removing entries. Due to the rather limited size and limited distinction between
 *  model and view, there is no division into two classes.
 *  @author Ralf Gitzel
 *  @version 1.0
 */
public class TableGUI extends JPanel implements ActionListener, AdjustmentListener, FocusListener
{

/** The column titles. The first one is the key by which table
 *  entries are identified.*/
private JLabel[] FieldNames;

/**  Which TableEntry-implementing object is to be used with
 *  the given column. */
private TableEntry[] NewLine;

/** The current table entries */
private Map Tabelle;

/** Contains TableZone and ScrollBar */
protected JPanel DisplayZone;

/** Here the table is displayed */
protected JPanel TableZone;

/** The scrollbar to move through the table */
protected JScrollBar Scroll;

/** The array of visible table components */
protected JComponent[][] VisiParts;

/** The current offset value */
private int Offset = 0;

/** The add/remove button bar */
protected JPanel ButtonBar;

/** The buttons to add a new entry */
protected JButton AddButton;

/** The last value of a field before focus was lost! */
private String OldVal;

/** The element that was the last one to gain focus! */
private TableEntry LastFocus;

/** The values in the new-line -> save when scrolling etc. */
private String[] CurrNewline;

private PropertyChangeSupport PListen;

/** This set contains all TableErrorListeners */
private Set TableErrLis;

/** The text to be written whenever a key is set to "" which must not happen!*/
private String KeyNull = "The key value must not be null!";

/** The text to be written when value for key is already in use! */
private String ValInUse = "This key value is already in use";

/** Defines a table that can be used to customize tables in objects visually.
 *  @param fieldNames The column titles. The first one is the key by which table
 *  entries are identified.
 *  @param entryTypes Which TableEntry-implementing object is to be used with
 *  the given column.
 *  @param rows The number of rows visible without scrolling. */
public TableGUI(String[] fieldNames, TableEntry[] entryTypes, int rows)
  {
  int i=0, j=0;
  PListen = new PropertyChangeSupport(this);
  TableErrLis = new HashSet();
  CurrNewline = new String[fieldNames.length];
  FieldNames = new JLabel[fieldNames.length];
  VisiParts = new JComponent[rows][fieldNames.length];
  for (i=0; i < fieldNames.length; i++)
    FieldNames[i] = new JLabel(fieldNames[i]);
  for (i=0; i < rows; i++)
    for (j=0; j < entryTypes.length; j++)
      {
      VisiParts[i][j] = (JComponent) entryTypes[j].otherInstance();
      }
  NewLine = new TableEntry[entryTypes.length];
  for(j=0; j < entryTypes.length; j++)
     NewLine[j] = entryTypes[j].otherInstance();
  DisplayZone = new JPanel(new BorderLayout());
  TableZone = new JPanel(new GridLayout(rows+1, entryTypes.length));
  Scroll = new JScrollBar(JScrollBar.VERTICAL);
  AddButton = new JButton("Add");
  ButtonBar = new JPanel();
  Tabelle = new HashMap();
  initDisplay();
  }

/** Fill the table with information from the underlying data structure. This
 *  is a shortcut for using addTableEntry() inputting data in a rawer form. Note
 *  that this fires PropertyChangeEvents!
 *  @param data first index is the row number, second one the column. */
public void setTable(String[][] data)
  {
  Tabelle.clear(); // remove old stuff first
  String[] line = new String[data[0].length-1];  // line minus key value
  int j=0;
  for (int i=0; i < data.length; i++)
    {
    for (j=0; j < line.length; j++)
      line[j] = data[i][j+1];
    setTableEntry(data[i][0], line);
    }
  }

/** Read the data contained in the table. In a way this can be seen as a short
 *  form for getTableEntry(). */
public String[][] getTable()
  {
  int i = 0, j = 0;
  Object[] temp = Tabelle.keySet().toArray(); // all keys as strings
  String key = null; // the current key value
  String[] data = null; // the corresponding data fields
  String[][] out = new String[Tabelle.size()][FieldNames.length]; // the return value
  for (i=0; i < Tabelle.size(); i++)
    {
    key = (String) temp[i];
    out[i][0] = key;
    data = getTableEntry(key);
    for (j=1; j < FieldNames.length; j++)
      out[i][j] = data[j-1];
    }
  return out;
  }

/** Changes a table entry to the given values. Note that this is a bound property.
 *  The values (old and new) take the form of String arrays where the first entry
 *  is the key and the rest are the data part. If a key is removed (see removeTableEntry())
 *  the new value will be null! */
public void setTableEntry(String key, String[] data)
   {
   if (data.length != FieldNames.length-1)
     throw new IndexOutOfBoundsException(
         "Parameter 'data' must be String["+FieldNames.length+"], not size "+data.length+".");
   // compute new value for propertyChange
   String[] neu = new String[FieldNames.length];
   neu[0] = key;
   for(int i=1; i < FieldNames.length; i++)
     {
     neu[i] = data[i-1];
     }
   // compute old value if possible (i.e. != null)
   String[] old = getTableEntry(key);
   String[] alt = null;
   if (old != null)
     {
     alt = new String[FieldNames.length];
     alt[0] = key;
     for(int i=1; i < alt.length; i++)
       {
       alt[i] = old[i-1];
       }
     }
   Tabelle.put(key, data);
   PListen.firePropertyChange("TableEntry", alt, neu);
   }

/** See description of setTableEntry for details */
public void removeTableEntry(String key)
   {
   String[] old = (String[]) Tabelle.get(key);
   String[] alt = new String[old.length+1];
   alt[0] = key;
   for(int i=1; i < alt.length; i++)
     {
     alt[i] = old[i-1];
     }
   Tabelle.remove(key);
   PListen.firePropertyChange("TableEntry", alt, null);
   }

/** Returns a table entry. */
public String[] getTableEntry(String key)
   {
   return (String[]) Tabelle.get(key);
   }

/** Adds a propertyChangeListener. The only bound property is TableEntry, i.e.
 *  a propertyChange occurs every time a table entry is set. If there was an old
 *  one with the same key we get the old value, otherwise it is null! */
public void addPropertyChangeListener(PropertyChangeListener lis)
  {
  PListen.addPropertyChangeListener(lis);
  }

public void removePropertyChangeListener(PropertyChangeListener lis)
  {
  PListen.removePropertyChangeListener(lis);
  }

/** Set up the display to make it suitable for the table. */
private void initDisplay()
  {
  int i=0, j=0;
  // basic setup
  add(DisplayZone, BorderLayout.CENTER);
  add(ButtonBar, BorderLayout.SOUTH);
  // ButtonBar
  ButtonBar.removeAll();
  ButtonBar.add(AddButton);
  AddButton.addActionListener(this);
  // the table
  DisplayZone.removeAll();
  DisplayZone.setLayout(new BorderLayout());
  DisplayZone.add(Scroll, BorderLayout.EAST);
  Scroll.addAdjustmentListener(this);
  DisplayZone.add(TableZone, BorderLayout.CENTER);
  for (i = 0; i < VisiParts.length; i++)
    for (j = 0; j < VisiParts[i].length;j++)
      {
      TableZone.add(VisiParts[i][j]);
      VisiParts[i][j].addFocusListener(this);
      }
  updateLines();
  }

protected void updateLines()
  {
  // adjust scroll bar
  Scroll.setMaximum(Tabelle.size());
  // normal update
  int j = 0;
  String[] keyz;
  Object[] tmpO = Tabelle.keySet().toArray();  // Note: tempO with letter O not zero!!!
  keyz = new String[tmpO.length];
  for (j=0; j<tmpO.length;j++) keyz[j] = (String) tmpO[j];
  String[] currData = null;
  /* We iterate through all visible fields and fill them with the appropriate
     data. The field after the last entry will be of a different colour and used
     to add new entries. All others behind that one will be disabled */
  for(int i = 0; i < VisiParts.length; i++)
    {
    if (i+Offset < keyz.length)
      {
      currData = getTableEntry(keyz[i+Offset]);
      for (j=0; j < VisiParts[i].length; j++)
        {
        // set to editable
        VisiParts[i][j].setEnabled(true);
        // fill with required content
        if (j == 0) ((TableEntry) VisiParts[i][j]).setText(keyz[i+Offset]);
        else ((TableEntry) VisiParts[i][j]).setText(currData[j-1]);
        // set proper color
        VisiParts[i][j].setBackground(SystemColor.window);
        }
       }
    if (i+Offset == keyz.length)
        {
        for (j=0; j < VisiParts[i].length; j++)
          {
          // set to editable (used for adding new entries)
          VisiParts[i][j].setEnabled(true);
          //text to fill in
          ((TableEntry) VisiParts[i][j]).setText(CurrNewline[j]);
          // setting color appropriate for new entry
          VisiParts[i][j].setBackground(SystemColor.textHighlight);
          }
        }
    if (i+Offset > keyz.length)
        {
        for (j=0; j < VisiParts[i].length; j++)
          {
          // set to not editable (this field only appears when low number of entries)
          VisiParts[i][j].setEnabled(false);
          // no text to fill in
          if (j == 0) ((TableEntry) VisiParts[i][j]).setText("n.a.");
          else ((TableEntry) VisiParts[i][j]).setText("");

          // setting color appropriate for uneditable fields
          VisiParts[i][j].setBackground(SystemColor.text);
          }
        }
    }
  }

////////////////// misc ////////////////////////////////////////////

/** Set the error message text to use when key value already taken. */
public void setValueInUseText(String text)
  {
  ValInUse = text;
  }

/** Set the error message text to use when key is empty */
public void setKeyNullText(String text)
  {
  KeyNull = text;
  }

/** Add a TableErrorListener to this object - these Listeners will be
 *  notified with an TableError Object whenever there is a wrong input. */
public void addTableErrorListener(TableErrorListener lis)
  {
  TableErrLis.add(lis);
  }

public void removeTableErrorListener(TableErrorListener lis)
  {
  TableErrLis.remove(lis);
  }

protected void fireTableError(String what)
  {
  TableError te = new TableError(what);
  Iterator it = TableErrLis.iterator();
  while(it.hasNext())
    {
    ((TableErrorListener) it.next()).tableErrorOccured(te);
    }
  }

/** Sets the text on the button that adds a  new table entry. Use this
 *  for localization */
public void setAddButtonText(String txt)
  {
  AddButton.setText(txt);
  }

/** Update the table according to the buttons pressed etc. While this might seem
 *  unclean (no proper model/view separation) the small scope might excuse this.
 */
public void actionPerformed(ActionEvent e)
  {
  // Case 1: the add button has been pressed....
  if (e.getSource() == AddButton)
    {
    // get pos of input line (last line)
    int curr = Tabelle.size() - Offset;
    // read values into text field
    String key = CurrNewline[0];
    String[] data = new String[FieldNames.length-1];
    for(int i=0; i < FieldNames.length-1; i++)
      data[i] = CurrNewline[i+1];
    // now try to insert new entry
    if (key.equals(""))
      {
      this.fireTableError(KeyNull);
      return;
      }
    while(true)
      {
      if (getTableEntry(key) != null)
        {
        // Oops - value already exists
        key = key + "'";
        }
      else
        {
        // value is save to add !
        setTableEntry(key, data);
        break;
        }
      }
    for (int x=0; x < CurrNewline.length; x++) CurrNewline[x] = "";
    updateLines();
    }
  }

/** This feature handles the scroll bar and the changes according to it */
public void adjustmentValueChanged(AdjustmentEvent e)
  {
  int i=0, j=0;
  if (LastFocus != null) // avoid info loss due to scrolling
    {
    updateTableEntry((JComponent) LastFocus);
    }
  Offset = e.getValue();
  updateLines();
  }

/** Update table entries when values are changed. Important: No lineUpdate is called
 *  here to avoid a reshuffle of the entries. */
public void focusLost(FocusEvent e)
  {
  System.out.println("FL:" + System.currentTimeMillis());
  // only table entries have any effect
  if (!(e.getSource() instanceof TableEntry)) return;
  updateTableEntry((JComponent) e.getSource());
  OldVal = null;
  }

/** Called by focusLost and its ilk to update table entries due to changes in
 *  text fields, scrolling etc. */
private void updateTableEntry(JComponent c)
  {
  int i = 0, j = 0;
  String[] data = new String[FieldNames.length-1];
  String key = null;
  // determine position
  outside: for (i=0; i <= VisiParts.length; i++)
    {
    for (j=0; j < VisiParts[i].length; j++)
      {
      if (c == VisiParts[i][j]) break outside;
      }
    }
  // the new entry field is to be saved separately
  if (i+Offset >= Tabelle.size())
    {
    for (int x=0; x < CurrNewline.length; x++)
      CurrNewline[x] = ((TableEntry) VisiParts[i][x]).getText();
    return;
    }
  // decide action according to position...
  for (int x=0; x < FieldNames.length-1; x++)
      data[x] = ((TableEntry) VisiParts[i][x+1]).getText();
  key     = ((TableEntry) VisiParts[i][0]).getText();
  switch(j)
    {
    case 0: // a key value! Remove old one! Add new one if unique!
    if (getTableEntry(key) != null)
      {
      // new value already taken!!!
      this.fireTableError(ValInUse);
      ((TableEntry) VisiParts[i][0]).setText(OldVal);
      }
    else removeTableEntry(OldVal);
    default:
    setTableEntry(key, data); // update new values!
    }
  }

/** If focus is gained we save the old value of that field in OldVal. */
public void focusGained(FocusEvent e)
  {
  System.out.println("FG:"+System.currentTimeMillis());
  if (!(e.getSource() instanceof TableEntry))
    {
    return;
    }
  OldVal = ((TableEntry) e.getSource()).getText();
  LastFocus = ((TableEntry) e.getSource());
  }


////////////////// interfaces and inner classes ////////////////////

/** Use this interface for classes that are special table entries */
public interface TableEntry
  {
  /** returns the current table entry as a string */
  public String getText();

  public void setText(String txt);

  /* Get an instance of this class with default content (i.e. either standard
     constructor or same constructor parameters as this object).
     This is a convenience method */
  public TableEntry otherInstance();

  }

/** Just a JTextField that implements TableEntry -> The standard access to
 *  table content. */
public static class StringEntry extends JTextField implements TableEntry
  {
  public TableEntry otherInstance()
    {
    return new StringEntry();
    }
  }

/** A table entry that is a JComboBox */
public static class ChoiceEntry extends JComboBox implements TableEntry
  {
  public ChoiceEntry(String[] values)
    {
    super();
    if (values != null) for (int i=0; i < values.length; i++)
      {
      this.addItem(values[i]);
      }
    }

  public String getText()
    {
    return (String) this.getSelectedItem();
    }

  public void setText(String t)
    {
    if (t == null || t.equals(""))
      {
      this.setSelectedIndex(0);
      System.out.println("Index to null!");
      }
    this.setSelectedItem(t);
    }

  public TableEntry otherInstance()
    {
    ChoiceEntry ret =  new ChoiceEntry(null);
    for (int i=0; i < this.getItemCount(); i++)
      ret.addItem(this.getItemAt(i));
    return ret;
    }
  }


}