
package viror.teach.anim;

import java.beans.PropertyChangeEvent;

/** This class can be attached to a motion path and arrive at
  its goal after several steps. Note that an acutal speed cannot
  be set which might seem like a loss at first but when one
  considers the fact that speed is usually only needed to
  either move the object the appropriate amount for the time segment
  or to calculate arrival time, it is obvious that no such
  measure is needed here and would be hard to calculate in any event
  considering the spline motion path. If this library is to be used
  for more physical animations the introduction of a new type of
  mover would have to be considered.*/
public class AnimActor extends AnimElement
{

public static int NORMAL  = 0;
public static int INVERSE = 1;


/** This value indicates how far the Actor has travelled on its
    path so far. Values between 0 and TravelTime are valid */
protected int Completion;

/** The number of "animation turns" it will take the Actor to
    arrive at the end of its motion pass. */
protected int TravelTime;

/** The value by which Completion changes evey animation cycle.
  This value should be 1 or -1 to give direction but may also
  take any other value */
protected int Step;

/** The value which has to be reached to change to state ARRIVED and
  stop moving. Note that if |Step| is not 1 we might never arrive
  since we simply pass by the goal state. */
protected int ArrivalState;

/** The actor waits this many animation cycles until it moves
  again. This value can be used to have an inital wait or to
  stop the actor in between */
protected int WaitTime;


/** The path this actor is attached to. You should atach an actor
    to only one path. If you want it to move on another path, remove
    it from the old one first. */
protected AnimPath Path;

/** Creates a circleshaped actor that arrives in ten turns.*/
public AnimActor()
    {
    super(VectorRep.createPrimitive(VectorRep.CIRCLE, 5), 0,0);
    TravelTime = 50;
    Step = 1;
    ArrivalState = TravelTime;
    }

/** Creates an actor.
  @param size The size in grid points
  @param shape Choose either VectorRep.CIRCLE or VectorRep.SQUARE
            as form
  @param traveltime The time in animation turns it takes the actor
            to arrive at its path's endpoint.
  @param direction Is either NORMAL to move from start to end or
            INVERSE to move from end to start. */
public AnimActor(int size, int shape, int traveltime, int direction)
    {
    super(VectorRep.createPrimitive(shape, size), 0,0);
    TravelTime = traveltime;
    if (direction == INVERSE)
      { Step = -1; ArrivalState = 0; Completion = traveltime; }
    if (direction == NORMAL)
      { Step =  1; ArrivalState = traveltime; Completion = 0; }
    }

/** Set the TravelTime. Changing Traveltime after the element has
    started t move will cause erratic behaviour (i.e. the next move
    will be a jump, possibly in the wrong direction) */
public void setTravelTime(int i)
    {
    TravelTime = i;
    }

/** Returns the TravelTime, i.e. how many turns it takes the Actor
    to move from the start of its motion path to the end */
public int getTravelTime()
    {
    return TravelTime;
    }

/** Returns the completion. Completion/TravelTime is the % value
    of how much of the path has been travelled yet. */
public int getCompletion()
    {
    return Completion;
    }

/** Changes the completion value of the actor. Be warned that this might
  not be the best idea after the actor has started moving. */
public void setCompletion(int c)
    {
    Completion = c;
    }

/** Sets the direction of the Actor when moving on the path. 1 means moving
  from 0 to TravelTime, -1 otherwise. Note that if travelling backwards
  it is important to remeber that Completion will be counted down not
  up. In other words, step is the value that increments Completion every turn.
  */
public void setStep(int i)
    {
    Step = i;
    }

/** Returns the value that is added to Completion every turn. */
public int getStep()
    {
    return Step;
    }

/** The value that must be reached to change state to ARRIVED */
public void setArrivalState(int i)
    {
    ArrivalState = i;
    }

/** The value that must be reached to change state to ARRIVED */
public int getArrivalState()
    {
    return ArrivalState;
    }

/** Sets the waiting time */
public void setWaitTime(int i)
    {
    WaitTime = i;
    }

/** The time the actor waits before moving again (not between each step
  but rather before the next one) */
public int getWaitTime()
    {
    return WaitTime;
    }

public void propertyChange(PropertyChangeEvent e)
    {
    }

public void move()
    {
    if (Path == null) return; // no path - no movement

    if (WaitTime > 0)
      {
      WaitTime--;
      return;   // only move, after the break has ellapsed
      }

    // check if we still have to move
    if (Completion == ArrivalState)
      {
      return; // we are already at the destination!
      }
    Completion += Step; // moving one up
    // recalculate position
    double percent = ((double) Completion) / ((double) TravelTime);
    setGridPosition(Path.getActorPosition(percent));
    // check if we are at the end (and notify listeners)
    if (Completion == ArrivalState)
      {
	setState(ARRIVED);
	if ((getState()&SELECTED) != 0) setState(SELECTED); // unselect
      }
    }

/** Do not call this method. It is to be used exclusively by AnimPath
    when it adds an actor. To add an actor to a path call:
    AnimPath.addActor() instead. */
boolean setPath(AnimPath p)
    {
    /* Are our conditions met? We either have to have no path set yet
       or if there is one already, we "delete" it by setting it to null */
    if (Path != null && p != null) return false;
    Path = p;
    return true;
    }

/** returns the path this element moves on */
public AnimPath getPath()
    {
    return Path;
    }

/** When the actor is copied the state is reset to NOSTATE */
public Object clone()
    {
    AnimActor a = (AnimActor) super.clone();
    a.Path = null;
    a.setState(getState());
    return a;
    }

}
