
//Titel:Viror Animationsbibliothek
//Version:
//Copyright:Copyright (c) 2000
//Autor:Ralf Gitzel
//Organisation:
//Beschreibung:Diese Sammlung von JavaBeans liefert eine grundlegende Bibliotek
//zur Erstellung von TeachWare, welche Animationen verwendet.

package viror.teach.anim;

import java.beans.*;
import java.awt.event.*;
import java.awt.Dimension;
import javax.swing.*;

public abstract class AnimElement implements PropertyChangeListener,
                                             MouseListener,
                                             MouseMotionListener,
                                             Cloneable
{

  /** For Details on the various states see the diagram "Elementstates" */
  private int State;

  /** The default state of the element. */
  public final static int NOSTATE  = 0;

  /** This flag is set when the element has arrived at the end of its
    path (AnimActors) or more generally speaking has reached the
    end of its activity (e.g. a derieved class "Mixer" in an industrial
    simulation could change state to ARRIVED after finishing its work) */
  public final static int ARRIVED  = 1;

  /** When an object is removed from the Animation it chnages to this
    state to give all listeners the chance to react to this. */
  public final static int DELETED  = 2;

  /** This flag is set when the object is selected which happens after
    the mouse is clicked on it but may also be activated directly. */
  public final static int SELECTED = 4;

  protected transient PropertyChangeSupport propertyChangeListeners = new PropertyChangeSupport(this);

  /** The x coordinate in the "fictive" grid */
  protected int GridX;

  /** The y coordinate in the "fictive" grid */
  protected int GridY;

  /** The physical representation of this component */
  private AnimRepresentation Peer;

  /** The AnimList where the Element is currently stored. Having an element
    in two lists is not possible (and usually not desireable) */
  private AnimList AList;

  /** This value is used to check whether the user can drag the component
      with the mouse. Default is false and Actors and Paths may not change
      this. */
  protected boolean Dragable;

  /** The identification string for this AnimElement. */
  private String Identification;


  /** This constructor is to be used by all derieved classes to avoid
    the evil of nullpointer exceptions. peer may be null. */
  AnimElement(AnimRepresentation peer, int x, int y)
    {
      if (peer != null)
	{
	  setPeer(peer);
	}
      setGridPosition(x,y);
    }


  public int getState()
  {
    return State;
  }

  /** Sets the state to this value. Note that it will be ANDed to the
      currently existing states. If newState is already set (i.e.
      newState AND State = newState), it will be unset by this call.
      In other words, if an element is ARRIVED and you want to undo this,
      call setState(ARRIVED) */
  public synchronized void setState(int newState)
  {
    int  oldState = State;
    State = State ^ newState;
    propertyChangeListeners.firePropertyChange("State",
					       new Integer(oldState),
					       new Integer(State));
    // check if the state change requires any other changes.
    if ((newState & SELECTED) != 0)
      Peer.setHighlighted(!Peer.isHighlighted());
  }

  public synchronized void removePropertyChangeListener(PropertyChangeListener l)
  {
    propertyChangeListeners.removePropertyChangeListener(l);
  }

  public synchronized void addPropertyChangeListener(PropertyChangeListener l)
  {
    propertyChangeListeners.addPropertyChangeListener(l);
    // the new listener gets the initial values
    Dimension pos = new Dimension(GridX, GridY);
    Integer s = new Integer(State);
    propertyChangeListeners.firePropertyChange("GridPosition", pos, pos);
    propertyChangeListeners.firePropertyChange("State", s, s);
  }

/** Sets the AnimRepresentation which should be used with this element.
 *  If using BitmapReps the InitialScale will not be passed on, i.e. the new
 *  one must have its InitialScale set individually.
 *  IMPLEMENTATION NOTE: if there already was a peer, special care has to be
 *  taken; we have to initialize the peer the same way we would when an element
 *  is added to an AnimList. Also the old peer has to be removed as a property
 *  change listener (to make it garbage collectible!) */
public void setPeer(AnimRepresentation newPeer)
  {
    if (Peer != null)
      {
      removePropertyChangeListener(Peer);
      if (Peer.getDisplay() != null)
        {
        newPeer.setDisplay(Peer.getDisplay());
        newPeer.scale(Peer.getDisplay().getScale());
        newPeer.setPosition(getGridPosition());  // 'fire property Change'
        }
      }
    Peer = newPeer;
    addPropertyChangeListener(Peer);
  }

  /** returns the peer element. Be aware that this might very well be null! */
  public AnimRepresentation getPeer()
  {
  return Peer;
  }

  public boolean isDragable()
  {
    return Dragable;
  }


  /** What happens in this function depends on the type of
      Element. Static one for example will do nothing. */
  public abstract void move();

  /** called when the element is deleted. Again depends on type as
      different elements have different relations and the delete
      depends might affect others (i.e. actors moving along a path).
      super.remove() should be called however. */
  public void remove()
    {
      setState(DELETED);
    }

  /** Mouse events occuring inside this elements representation
      will be handled here (in the case of a click)
      In the default a click will select an element but
      by overwriting this function, more exciting things can be
      implemented (image maps, switches etc.) */
  public void mouseClicked(MouseEvent e)
  {
    setState(SELECTED);
  }

  public void mousePressed(MouseEvent e)
    {
      // do nothing - but derieved classes may...
    }

  public void mouseReleased(MouseEvent e)
    {
      // do nothing - but derieved classes may...
    }

  public void mouseEntered(MouseEvent e)
    {
      // do nothing - but derieved classes may...
    }

  public void mouseExited(MouseEvent e)
    {
      // do nothing - but derieved classes may...
    }

  public void mouseMoved(java.awt.event.MouseEvent e)
    {
    }

  /** AnimElements may be dragged over the display if they are
    dragable. */
  public void mouseDragged(MouseEvent e)
    {
    if (!Dragable)
      {
	return;
      }
    Dimension temp = new Dimension(e.getX(), e.getY());
    setGridPosition(temp);
    }

  /** when we get a DELETED from any other object we sever the ties to
    it -> this ensures that the object will be gone completely from
    this animation and if not used elsewhere will be consumed by
    the garbage collector daemon. */
  public void propertyChange(PropertyChangeEvent e)
    {
    /* On this stage we ignore everything that is not an AnimElement */
    AnimElement elem = null;
    try {
      elem = (AnimElement) e.getSource();
        } catch (Exception exep) { return; }
    // sever ties to deleted element
    if (e.getPropertyName().equals("State"))
	{
	int val = 0;
	try {
	  val = ((Integer) e.getNewValue()).intValue();
	    } catch (Exception exe)
	       { System.err.println("State is int only!"); }
	if ((val & AnimElement.DELETED) != 0)
	  {
	  elem.removePropertyChangeListener(this);
	  }
	}
    }



  /** Sets the grid position (i.e. the coordinates based on the
      'fictive' grid and not the actual display size). */
  void setGridPosition(int x, int y)
    {
    Dimension otmp = new Dimension(GridX, GridY);
    GridX = x;
    GridY = y;
    Dimension tmp = new Dimension(x,y);
    propertyChangeListeners.firePropertyChange("GridPosition", otmp, tmp);
    }

  /** Sets the grid position (i.e. the coordinates based on the
      'fictive' grid and not the actual display size). */
  void setGridPosition(Dimension d)
    {
    setGridPosition(d.width, d.height);
    }

  /** returns the grid position */
  Dimension getGridPosition()
    {
    return new Dimension(GridX, GridY);
    }

  /** This function should only be called by an implementation of
    AnimList during add and remove. Other usage is NOT recommended */
  protected void setList(AnimList l)
    {
    AList = l;
    }

  /** Returns the AnimList this element is stored in */
  public AnimList getList()
    {
    return AList;
    }

  /** Returns a clone of this object. The copy of the peer is deep. The
   PropertyChangeListeners are not copied. */
  public Object clone()
    {
    AnimElement ret = null;
    try{
      ret = (AnimElement) super.clone();
      ret.propertyChangeListeners = new PropertyChangeSupport(ret);
      if (Peer != null)
	ret.setPeer((AnimRepresentation) this.getPeer().clone());
    } catch (Exception e) { System.err.println("AnimElement: No Clone!"); }
    return ret;
    }

  /** The identification of this AnimElement. This is not an individual name
   *  but rather a group identification. This differs from class - elements of
   *  different classes may belong to the same identification-"group". */
  public String getIdentification()
    {
    return Identification;
    }

   /** The identification of this AnimElement. */
  public void setIdentification(String id)
    {
    Identification = id;
    }

/** This is a convenience method - since most elements use bitmaps
 *  as representations. It scales the BitmapRep peer according to user given value.
 *  All other reps will be ignored by this method. */
public void scalePeer(double val)
  {
  BitmapRep r = null;
  try
    {
    r = (BitmapRep) getPeer();
    r.setIniScale(val);
    } catch (ClassCastException e)
      {
      System.err.println("Trying to scale non-bitmap peer in "+this+". Command ignored!");
      }
  }


  /** Returns a beans customizer that either editable or not.
    By default null is returned but derieved classes might very
    well have one that will be used by StandardButtons. Please
    note that each known implementation of this returns a <i>new</i>
    object with each call, so if multiple ops are to be performed
    it is advisebale to call this mehtod only once!
    All Objects returned here implement the UndoCustomizer-Interface!*/
  public JComponent getCustomizer(boolean editable)
    {
    return null;
    }

}









