
package viror.teach.anim;

import java.awt.Dimension;
import java.awt.Image;
import java.util.*;
import java.beans.PropertyChangeEvent;

/** This class is a motion path that may have actors attached
    to it. Pathes are simplex, i.e. you can only move from
    start to end and not the other way round. It is however
    possible to have two pathes with the same spline only that
    the control points go in different directions. The shape
    of the curve is controlled via the tangents (see setStart/EndTangent).
    To create more complex curves, several paths have to be combined
    via 'cloaked' endpoints.
    Note that while bitmaps and polygons are valid represenations for the
    AnimPath, their generic nature does not allow them to to be
    updated when endpoints are dragged, so these features should be
    used carefully. */
public class AnimPath extends AnimElement
{

/** The starting point of the Path */
private AnimStatic Start;

/** The end point of the path */
private AnimStatic End;

/** The set of all ids allowed for the startpoint */
protected Set StartSet = null;

/** The set of all ids allowed for the endpoint */
protected Set EndSet = null;

/** The list of actors on this path */
protected Set Actor;

/** The acutal path */
protected Spline Path;

/** For more info on this see generateRepresentation */
protected boolean OpticUpdate;

/** The standard constructor. Usually the Peer is created afterwards
 with generateRepresentation(). */
public AnimPath()
    {
    super(null, 0,0);
    Actor = new HashSet();
    Path = new Spline();
    OpticUpdate = false;
    }

/** This constructor creates a path that has an image as peer.
  While unusual some cases may warrant this. Note that the start and endpoint
  still have to be set and if they are not at exactly the right distance
  things will look odd.
  @param img The image to represent the path.
  @param sel The image to be used when the path is selected */
public AnimPath(Image img, Image sel)
    {
    super (new BitmapRep(img, sel), 0,0);
    Actor = new HashSet();
    Path = new Spline();
    OpticUpdate = false;
    }

/** Sets the beginning of the path which has to be a static
    element. Note that the startspoint is also what we get returned
    when we call getGridPosition. While setGridposition exists in
    this class as well, its use is highly discouraged after setStart
    has been called (in fact if a start has been specified, it will have
    no effect) */
public void setStart(AnimStatic start) throws IllegalArgumentException
    {
    // check whether this point is allowed.
    if (StartSet != null && !StartSet.contains(start.getIdentification()))
       throw new IllegalArgumentException(start + " is not allowed to connect to" +this);
    // compute current tangent offset
    Dimension tan = Path.getControlPoint(1); // get abs value
    Dimension p; // here the old startpoint comes in to get relative value
    if (Start != null) p = Path.getControlPoint(0);
    else p = new Dimension();
    int x = tan.width  - p.width;
    int y = tan.height - p.height;
    // remove old start point
    if (Start != null)
      {
      removePropertyChangeListener(Start);
      Start.removePropertyChangeListener(this);
      Start.disconnect(this); // inform startpoint that it has been dismissed
      }
    Start = start;
    Start.connect(this); // inform new startpoint that it was connected
    addPropertyChangeListener(Start);
    Start.addPropertyChangeListener(this);
    Dimension d = Start.getGridPosition();
    Path.setControlPoint(0,d);
    // also update the tangent
    setStartTangent(x,y);
    // set grid position required for updates etc.
    super.setGridPosition(d.width, d.height);
    }

public AnimStatic getStart()
    {
    return Start;
    }

/** Sets the end point of the path.
 */
public void setEnd(AnimStatic end)
    {
    // check whether this point is allowed.
    if (EndSet != null && !EndSet.contains(end.getIdentification()))
       throw new IllegalArgumentException(end + "is not allowed to connect to" +this);
    // compute current tangent offset
    Dimension tan = Path.getControlPoint(2); // get abs value
    Dimension p; // here the old startpoint comes in to get relative value
    if (End != null) p = Path.getControlPoint(3);
    else p = new Dimension(0,0);
    int x = tan.width  - p.width;
    int y = tan.height - p.height;
    // remove old EndPoint
    if (End != null)
      {
      removePropertyChangeListener(End);
      End.removePropertyChangeListener(this);
      End.disconnect(this); // tell old end that it was dismissed
      }
    End = end;
    End.connect(this); // tell new end that it is now connected..
    addPropertyChangeListener(End);
    End.addPropertyChangeListener(this);
    Dimension d = End.getGridPosition();
    Path.setControlPoint(3,d);
    // also update the Tangent
    setEndTangent(x,y);
    }

public AnimStatic getEnd()
    {
    return End;
    }

/** Only call this function if no startpoint has been set yet.
  @see setStart */
public void setGridPosition(int x, int y)
    {
    if (Start != null) return;
    super.setGridPosition(x,y);
    }

/** Adds an actor to the motion path.
    @param actor The actor to add to the list.
    @throws IllegalStateException If the actor is already in the list
        of this path we cannot add it again and get this exception. */
public void addActor(AnimActor actor) throws IllegalStateException
    {
    // first check if this actor is already on another path
    if (!actor.setPath(this))
      {
      throw new IllegalStateException("Actor already on path!");
      }
    Actor.add(actor);
    actor.addPropertyChangeListener(this); // to be notified of arrival
    // move to path start position
    double per = ((double) actor.getCompletion())
                 / ((double) actor.getTravelTime());
    actor.setGridPosition(getActorPosition(per));
    }

/** Removes an actor from the list of elements moving along the
  path. */
public boolean removeActor(AnimActor actor)
    {
    boolean wasThere = Actor.remove(actor);
    if (wasThere)
      {
	actor.removePropertyChangeListener(this);
	actor.setPath(null);
      }
    return wasThere;
    }

/** The curve used for the Animpath is a cubic spline. This is not unlike
  the curve segment between two points as seen in products like corel draw.
  We have tangents in the endpoints that we can move, thus changing the
  shape of the curve. The longer the tangent gets the bigger the arc of the
  curve becomes. Tangent coordinates are given relative to the endpoint
  they belong to.
  @param xoff Offset to the x coordinate of the startpoint
  @param yoff Offset to the y coordinate of the startpoint*/
public void setStartTangent(int xoff, int yoff)
    {
    Dimension set = new Dimension();
    Dimension start = Path.getControlPoint(0);
    set.width  = start.width  + xoff;
    set.height = start.height + yoff;
    Path.setControlPoint(1,set);
    }

/** Passes a set of all Identifications that will be allowed as start points. If
 *  this set is null no id-check will be performed. Otherwise all ids not listed
 *  in the set will cause an IllegalArgumentException.
 *  <p><b>Common Error:</b> If Start and End are supposed to have the same
 *  sets allowed but these should change over time it is a bad idea to pass the
 *  same set -> all alterations will affect both start and end since we have the
 *  same object with two references. */
public void setAllowedStartPoints(Set s)
   {
   StartSet = s;
   }

/** Adds a single Identification to the AllowedStartPoints. */
public void addAllowedStartPoint(String id)
   {
   if (StartSet == null) StartSet = new HashSet();
   StartSet.add(id);
   }

/** Removes a single Identification from the AllowedStartPoints */
public void removeAllowedStartPoint(String id)
  {
  StartSet.remove(id);
  }

/** see the other form of setStartTangent for details */
public void setStartTangent(Dimension d)
    {
    setStartTangent(d.width, d.height);
    }

/** The curve used for the Animpath is a cubic spline. This is not unlike
  the curve segment between two points as seen in products like corel draw.
  We have tangents in the endpoints that we can move, thus changing the
  shape of the curve. The longer the tangent gets the bigger the arc of the
  curve becomes. Tangent coordinates are given relative to the endpoint
  they belong to. (Note that you will probably want the offsets for the
  endpoint tangent to be negative to get good results)
  @param xoff Offset to the x coordinate of the endpoint
  @param yoff Offset to the y coordinate of the endpoint*/
public void setEndTangent(int xoff, int yoff)
    {
    Dimension set = new Dimension();
    Dimension end = Path.getControlPoint(3);
    set.width  = end.width  + xoff;
    set.height = end.height + yoff;
    Path.setControlPoint(2,set);
    }

/** see the other form of setEndTangent for details */
public void setEndTangent(Dimension d)
    {
    setEndTangent(d.width, d.height);
    }

/** Passes a set of all Identifications that will be allowed as end points. If
 *  this set is null no id-check will be performed. Otherwise all ids not listed
 *  in the set will cause an IllegalArgumentException.
 *  <p><b>Common Error:</b> If Start and End are supposed to have the same
 *  sets allowed but these should change over time it is a bad idea to pass the
 *  same set -> all alterations will affect both start and end since we have the
 *  same object with two references. */
public void setAllowedEndPoints(Set s)
   {
   EndSet = s;
   }

/** Adds a single Identification to the AllowedEndPoints. */
public void addAllowedEndPoint(String id)
   {
   if (EndSet == null) EndSet = new HashSet();
   EndSet.add(id);
   }

/** Removes a single Identification from the AllowedEndPoints */
public void removeAllowedEndPoint(String id)
  {
  EndSet.remove(id);
  }

/** sets the peer to a spline that looks just like the one we use
  to send our actors along. Utilisation of this function is highly
  recommended as it:
  <ol>
    <li>Guarantees that actors will move along the drawn path
    <li>If endpoints change their position, the optical representation
        will automatically adapt. (This is done internally by setting
	the OpticUpdate flag to true)
  </ol>
  In some rare cases one might want to use a bitmap etc. for the path -
  setPeer can be used than instead of this method but one has to be
  aware that endpoint changes have to be handled seperately (best done
  with a propertyChangeListener to the end points). */
public void generateRepresentation()
    {
    setPeer(new SplineRep());
    alterRepresentation();
    }


/** this function updates the Peer. It will fail if the Peer is not
  of type SplineRep. Note that peer update only works for splines
  of degree 3.
  @throws UnsupportedOperationException when trying to call this function
      for a path that does not have a spline attached, we will get
      this exception. It will also occur if we try to update the
      curve and have a motion path spline of order higher than 3. */
public void alterRepresentation() throws UnsupportedOperationException
    {
      SplineRep s = null;
      Dimension temp;
      try{
	s = (SplineRep) getPeer();
      } catch( Exception e)
	{
	  throw new UnsupportedOperationException(
	   "False type of rep for alterRepresentation!");
	}
      if (Path.getDegree() != 3)
	throw new UnsupportedOperationException(
	  "Can only autogenerate Splines of degree 3!");
      s.setToSpline(Path);
    }

/** There are two kinds of propertyChanges that interest us. First when
  the endpoints move, we update our Motion path spline and <i>maybe</i>
  its optical representation (according to user setting - see
  generateRepresentation for details). Second we want to react to actors
  arriving (in the basic package we do nothing overly exciting in that
  case other than removing the actor - inheriting classes will probably
  overwrite the protected function handleArrival) */
public void propertyChange(PropertyChangeEvent e)
    {
      super.propertyChange(e);
      // case start/endpoint changed
      if (e.getPropertyName().equals("GridPosition"))
	{
	  AnimStatic s = null;
	  // the try/catch block ensures that only AnimStatics are regarded
	  try {
	  s = (AnimStatic) e.getSource();
	  } catch (Exception ex) { return; }
	// check if address of source and start point is the same
	if (s == Start)
	  {
	  setStart(s);
	  }
	if (s == End)
	  {
	  setEnd(s);
	  }

	// change path and move actors to new position
	try {
	alterRepresentation();
	} catch (Exception ex) {} // only if generated path
	Iterator it = Actor.iterator();
	AnimActor curr = null;
	while (it.hasNext())
	  {
	  curr = (AnimActor) it.next();
	  curr.setGridPosition(getActorPosition(1.0 * curr.getCompletion() /
						1.0 * curr.getTravelTime()));
	  }
	return;
	}

      // case state change
      AnimElement source = null;
      try {
	source = (AnimElement) e.getSource();
      } catch (ClassCastException ex)
	{
	System.err.println("Property 'State' reserved for AnimElements");
	}
      if (e.getPropertyName().equals("State"))
	{
	int newval = ((Integer) e.getNewValue()).intValue();

	// case1: Actor arrives
	if ((newval & AnimElement.ARRIVED) != 0)
	  {
	  // this MUST be an actor, others never attain this state
	  handleArrival((AnimActor) e.getSource());
	  }
	// case2: element was removed
	if ((newval & AnimElement.DELETED) != 0)
	  {
	  // case actor
	  if (source instanceof AnimActor)
	    {
	    removeActor((AnimActor) source);
	    }
	  if (source instanceof AnimStatic)
	    {
	    //this MIGHT be one of our endpoints
	    if (source == Start | source == End)
	      {
	      // we go down with it
	      getList().removeElement(this);
	      }
	    }
	  }
	}
    }


/** This method is called whenever an AnimActor arrives at its goal.
  The basic animation does nothing but derieved classes will probably
  want to react to this event and will override this method. */
protected void handleArrival(AnimActor e)
    {
    }

/** returns the position an actor should have after
    progressing percentage% on the motion path */
public Dimension getActorPosition(double percentage)
    {
    return Path.computePoint(percentage);
    }

/** like statics we do not move */
public void move()
    {
      // do nothing
    }

/** If the AnimPath is removed all Actors on it have to go as well */
public void remove()
    {
    super.remove();
    AnimActor ac = null;
    // kill all actors
    Iterator it = Actor.iterator();
    while (it.hasNext())
      {
      ac = (AnimActor) it.next();
      getList().removeElement(ac);
      }
    }

  /** Generate a copy of this path but only its "look and feel". The
    Start and End are set to null, the ActorList is reinitialized and
    the internal Spline copied. */
  public Object clone()
    {
    AnimPath p = (AnimPath) super.clone();
    p.Start = null;
    p.End = null;
    p.Actor = new HashSet();
    p.Path = new Spline();
    return p;
    }

}
