
//Titel:        Viror Animationsbibliothek
//Version:
//Copyright:    Copyright (c) 1998
//Autor:       Ralf Gitzel
//Firma:
//Beschreibung:  Diese Sammlung von JavaBeans liefert eine grundlegende Bibliothek
//zur Erstellung von TeachWare, welche Animationen verwendet.

package viror.teach.anim;

import java.util.*;
import java.awt.Dimension;
import java.awt.Cursor;
import java.awt.event.*;
import java.awt.Point;

/** This is a standard implementation for AnimList which seems
    adequate in most cases. It stores the different layers (represented
    by HashSets) in a List (in this case an ArrayList to ensure constant
    access time). GetPeers merely iterates over each HashSet but orders
    the layers such that our paint actions will provide the proper layering
    (i.e. lowest objects are painted first).
    For convenience sake, we make SimpleAnimList a KeyListener. It can be
    attached to the AnimDisplay or surrounding frame (this has to be done
    seperately) and will delete the currently selected element if the
    backspace or delete key is pressed. If you do not wish to use these
    features just leave the List unattached in its key listener function. */
public class SimpleAnimList extends Thread implements AnimList,
                                               KeyListener, AnimCycle
{


/** The data structure used to store the different layers. Each layer
    is in itself a collection of the objects stored within that layer. */
private List Layer;

/** A reference back to the display for scaling etc. This should never be
    set manually, rather AnimDisplay manipulates this one when the list
    is passed. */
private AnimDisplay Display;

/** The state the list is in */
private int Mood;

/** The object that will be cloned to be inserted when in insert mode */
private AnimElement Insertor;

/** The pause value is used to stop/resume the thread for animation */
private Communicate Pause = new Communicate();

/** The time that elapses between two frames. */
private int WaitTime = 200;


/** This variable is used for dragging. mousePressed picks an element
  for dragging, mouseDragged moves this element and mouseReleased resets
  it to null. Reasons for this implementation are given in the manual. */
private AnimElement DragParty;

/** The last element that was entered. Used to handle MouseEntered and
  MouseExited events in the elements */
private AnimElement InsideParty;

/** The first element of a manual connect */
private AnimStatic ConnectParty;

/** The path used to connect two elements. This variable is used when
  we go into mood CONNECT_ONE. */
AnimPath ConnectPath;

/** Will be true if our internal garbage collection (doGC-Method) will
  have to delete elements that are in state DELETED. */
boolean DeleteChange;

/** Will be true if our internal insertion list has to be merged with
  our data structure (i.e. elements were inserted).
  @see doIC() */
boolean InsertChange;

/** The insertion list - these elements will be inserted into the
  data structure. */
List InsertList;

public SimpleAnimList()
  {
    Layer      = new ArrayList();  // changes infrequently
    InsertList = new ArrayList();
  }

public synchronized AnimRepresentation[] getPeers()
  {
  doGC();
  doIC();
  AnimRepresentation[] retval = new AnimRepresentation[getSize()];
  Iterator it = Layer.iterator();
  int counter = 0;
  while (it.hasNext())
    {
    Iterator innerit = ((Collection)it.next()).iterator();
    while (innerit.hasNext())
      {
      retval[counter] = ((AnimElement)(innerit.next())).getPeer();
      counter++;
      }
    }
  return retval;
  }

public synchronized void moveAll()
  {
    doGC();
    doIC();
    AnimElement currel = null;
    Collection curr = null;
    Iterator it = Layer.iterator();  // iterator over the layers
    while (it.hasNext())
      {
      curr = (Collection) it.next(); // get current layer
      Iterator currit = curr.iterator();
      while (currit.hasNext())
	{
	currel = (AnimElement) currit.next();
	currel.move();
	}
      }
  }

public synchronized void addElement(int i, AnimElement e)
    {
    SALNewElement newElem = new SALNewElement();
    newElem.Element = e;
    newElem.Layer = i;
    InsertList.add(newElem);
    InsertChange = true;
    // set return reference already
    e.setList(this);
    }

/** Implementation note: If trying to remove an element that
  has just been inserted the deletion might not work. This
  problem only occurs if the add/remove-combo happens
  before the element was even shown on screen. */
public synchronized boolean removeElement(AnimElement e)
  {
  /* Note: we do not remove layers that become empty since
     they should not be a performance problem, unless somebody
     tries something rather odd like having 10000 layers.
     Also the Display variable of the peer has to be set to null*/
  boolean retval = false; // the return value
  boolean temp   = false; // temporary reference
  Iterator it = Layer.iterator();  // iterator over the layers
  while (it.hasNext())
    {
    Collection curr = (Collection) it.next(); // get current layer
    temp = curr.contains(e);
    if (temp == true) retval = temp;
    }

  if (!retval) return false; // this element was not in the list

  // remove this element from all others where it is a listener
  AnimElement pardner = null;
  it = Layer.iterator();
  while (it.hasNext())
    {
    Iterator innerit = ((Collection)it.next()).iterator();
    while (innerit.hasNext())
      {
      pardner = (AnimElement) innerit.next();
      pardner.removePropertyChangeListener(e);
      }
    }

  // call the "cleanup routine" and set state to deleted
  e.remove();

  // delete reference to this list
  e.setList(null);
  e.getPeer().Display = null;

  DeleteChange = true;

  /* NOT DELETED YET - the acutal deletion occurs when getAll() or
     getPeers() is called next time. If the animation is halted we
     call doGC() now to ensure that screen looks ok */
  try {
  if (Pause.isPause())
    {
      doGC();
      Display.repaint();
    }
  } catch (InterruptedException ex) {}
  return retval;
  }

public synchronized int getSize()
    {
    doGC();
    Iterator it = Layer.iterator();
    int count = 0;
    while (it.hasNext())
      {
      count += ((Collection) it.next()).size();
      }
    return count;
    }

public synchronized AnimElement[] getAll()
    {
    doGC();
    doIC();
    AnimElement[] retval = new AnimElement[getSize()];
    Iterator it = Layer.iterator();
    int counter = 0;
    while (it.hasNext())
      {
      Iterator innerit = ((Collection)it.next()).iterator();
      while (innerit.hasNext())
	{
	retval[counter] = (AnimElement)(innerit.next());
	counter++;
	}
      }
    return retval;
    }

public synchronized AnimElement[] getAllSelected()
    {
    doGC();
    AnimElement[] retval = new AnimElement[getSize()];
    AnimElement curr = null;
    Iterator it = Layer.iterator();
    int counter = 0;
    while (it.hasNext())
      {
      Iterator innerit = ((Collection)it.next()).iterator();
      while (innerit.hasNext())
	{
	curr = (AnimElement) innerit.next();
	if ((curr.getState() & AnimElement.SELECTED) != 0)
	  {
	  retval[counter] = curr;
	  counter++;
	  }
	}
      }
    AnimElement[] prunedret = new AnimElement[counter];
    for(int i=0; i<counter;i++) prunedret[i] = retval[i];
    return prunedret;
    }

public synchronized boolean removeAllSelected()
    {
    AnimElement[] all = getAllSelected();
    if (all.length == 0) return false;
    for(int i=0; i<all.length; i++) removeElement(all[i]);
    return true;
    }

public synchronized boolean removeAll()
    {
    AnimElement[] all = getAll();
    if (all.length == 0) return false;
    for(int i=0; i<all.length; i++) removeElement(all[i]);
    return true;
    }

  /** See AnimList for description. This version tests whether the
    AnimPath has a peer - if not it will be generated as a SplineRep
    via generateRepresentation. Note that if no objects where selected
    the connectSelected will set the list into rubberband connection mode. */
public synchronized boolean connectSelected(AnimPath p)
    {
    AnimElement[] all = getAllSelected();
    int anz = all.length;
    if (anz != 2)
      {
      setMode(CONNECT_ONE);
      ConnectPath = p;
      return true;
      }
    if (!(all[0] instanceof AnimStatic) || !(all[1] instanceof AnimStatic))
      return false;
    AnimPath p1 = (AnimPath) p.clone();
    p1.setStart((AnimStatic) all[0]);
    p1.setEnd((AnimStatic) all[1]);
    if (p.getPeer() == null)
      {
      // generate peers (will not be screen-based yet)
      p1.generateRepresentation();
      }
    addElement(0, p1);
    // adapt to screen resolution etc.
    p1.alterRepresentation();
    return true;
    }

public void setMode(int m)
    {
    Mood = m;
    if (Display == null) return;
    switch(m)
      {
      case INSERT:
      case CONNECT_TWO:
	Display.setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
      break;

      case NEUTRAL:
	Display.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
      break;

      case CONNECT_ONE:
	Display.setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
	break;
      }
    }

public int getMode()
    {
    return Mood;
    }

public void setInsertor(AnimElement e)
    {
    Insertor = e;
    }

/** read warning in the interface description! */
public void setDisplay(AnimDisplay d)
    {
    Display = d;
    AnimRepresentation[] all = getPeers();
    AnimElement[]       elem = getAll();
    for(int i=0; i < all.length; i++)
      {
	all[i].setDisplay(d);
	all[i].setPosition(elem[i].getGridPosition());
      }
    }

public AnimDisplay getDisplay()
    {
    return Display;
    }

public void initialize()
   {
     doGC();
     doIC();
     AnimElement[] all = getAll();
     for(int i=0; i< all.length; i++)
       {
       all[i].getPeer().scale(Display.getScale());
       all[i].getPeer().setPosition(all[i].getGridPosition());
       }
   }


/** The animation cycle. May be paused by setting Pause to true. */
public void run()
    {
    int nice = 0;
    setName("Animation-Thread");
    while(true)
      {
	moveAll();
	if (Display != null) Display.repaint();
	try {
	  Thread.sleep(WaitTime);
	  if (Pause.isPause())
	    {
	      // reserved for future use
	    }
	  nice++;
	  // We yield to ensure proper display in non-timeslice environments
	  if ( nice == 1000/ WaitTime) { nice=0; yield();}
	} catch (InterruptedException e)
	  { System.err.println("Interrupt!");}
      }
    }

public void suspendAnim()
    {
    Pause.setPause(true);
    }

public void resumeAnim()
    {
    Pause.setPause(false);
    }

public void setFrameRate(int i)
    {
    WaitTime = 100 / i;
    }

public int getFrameRate()
    {
    return (1000 / WaitTime);
    }

public AnimElement pick(Dimension point)
    {
    doIC();
    AnimElement picka = null;

    // 1. get a list of all elements
    AnimElement[] all = getAll();
    // 2. look at all in opposite painting order
    for (int i= all.length -1; i>=0; i--)
      {
      // fast check
      if (!all[i].getPeer().testBoundingBox(point)) continue;
      // accurate check
      if (!all[i].getPeer().testActual(point)) continue;
      // we got the good one!
      picka = all[i];
      break;
      }
    return picka;
    }


/** Sends a mouse event (clicked) to the AnimElement whose Peer
  was at the clicked position. Coordinates of the event are
  converted from actual to grid to allow the Element to handle
  them.
  Note that in the case of the Mood being INSERT, there will
  be no delegation, rather the mouse event will insert a copy of
  the current Insertor element.
  @see setInsertor
  @see setMood */
public void mouseClicked(MouseEvent e)
  {
  Dimension point = new Dimension((int) e.getPoint().getX(),
				  (int) e.getPoint().getY());
  // case 1, INSERT-Mood
  if (Mood == INSERT)
    {
    AnimElement neu = (AnimElement) Insertor.clone();
    addElement(Layer.size(), neu);
    neu.getPeer().scale(Display.getScale());
    neu.setGridPosition(Display.unconvertXY(e.getPoint()));
    return;
    }
  // case 2, NEUTRAL mood
  AnimElement pick = pick(point);
  if (pick == null)
    {
      return;
    }
  Point tmp = e.getPoint();
  Dimension temp = Display.unconvertXY(tmp);
  // set Event coordinates to grid (for use in AnimElement)
  e.translatePoint(-tmp.x + temp.width, -tmp.y + temp.height);
  pick.mouseClicked(e);
  }


/** A mousePressed-MouseEvent will have two effects. First
  the element inside of which the pressing occured will be
  notified with a mousePressed (event coordinates are adjusted
  to be relative to the element). Also we set the variable
  DragParty to the selected element to ensure proper dragging.
  If the mood is CONNECT_ONE we will switch to rubberband-mode
  now and go to CONNECT_TWO mood. */
public void mousePressed(MouseEvent e)
    {
    Dimension point = new Dimension((int) e.getPoint().getX(),
				    (int) e.getPoint().getY());
    AnimElement pick = pick(point);

    // case 1 - connect mode one
    if (getMode() == CONNECT_ONE)
      {
      // we try to connect but did not click on anything
      if (pick == null || !(pick instanceof AnimStatic))
	{
	setMode(NEUTRAL);
	return;
	}
      ConnectParty = (AnimStatic) pick; // we use this to store original node
      Display.activateRubberBand(e.getPoint());
      setMode(CONNECT_TWO);
      return;
      }

    // case 2 - other mode
    if (pick == null) return;
    Point tmp = e.getPoint();
    Dimension temp = Display.unconvertXY(tmp);
    // set Event coordinates to grid (for use in AnimElement)
    e.translatePoint(-tmp.x + temp.width, -tmp.y + temp.height);
    pick.mousePressed(e);
    DragParty = pick;
    }

/** Other than passing the event to the element we also set
  the variable DragParty to null. */
public void mouseReleased(MouseEvent e)
    {
    Dimension point = new Dimension((int) e.getPoint().getX(),
				    (int) e.getPoint().getY());
    AnimElement pick = pick(point);

    // case 1 - connect mode two
    if (getMode() == CONNECT_TWO)
      {
      if (pick == null)
	{
	setMode(NEUTRAL);
	Display.deactivateRubberBand();
        return;
	}
      AnimPath p1 = (AnimPath) ConnectPath.clone();
      AnimStatic EndP = null;
      try {
	EndP = (AnimStatic) pick;
      } catch (Exception exep)
	{
	setMode(NEUTRAL);
	Display.deactivateRubberBand();
	return;
	}
      try
        {
        p1.setStart(ConnectParty);
        p1.setEnd(EndP);
        if (p1.getPeer() == null)
	  {
	  // generate peers (will not be screen-based yet)
	  p1.generateRepresentation();
	  }
        addElement(0, p1);
        // adapt to screen resolution etc.
        p1.alterRepresentation();
        } catch(IllegalArgumentException ex)
           {
           // the connection is not performed...
           }
      setMode(NEUTRAL);
      Display.deactivateRubberBand();
      return;
      }

    // case 2 - other modes
    if (pick == null) return;
    Point tmp = e.getPoint();
    Dimension temp = Display.unconvertXY(tmp);
    // set Event coordinates to grid (for use in AnimElement)
    e.translatePoint(-tmp.x + temp.width, -tmp.y + temp.height);
    pick.mouseReleased(e);
    DragParty = null;
    }

public void mouseEntered(MouseEvent e)
    {
      // nothing
    }

public void mouseExited(MouseEvent e)
    {
      // nothing
    }

public void mouseMoved(MouseEvent e)
    {
    // may cause a mouse entered even in elements
    Dimension point = new Dimension((int) e.getPoint().getX(),
				    (int) e.getPoint().getY());
    AnimElement pick = pick(point);
    Dimension temp = Display.unconvertXY(e.getPoint());
    // set Event coordinates to grid (for use in AnimElement)
    e.translatePoint(-e.getPoint().x + temp.width, -e.getPoint().y +
		     temp.height);
    if (pick == null && InsideParty == null) return;
    if (pick == null && InsideParty != null)
      {
      // We have exited the Element
      InsideParty.mouseExited(e);
      InsideParty = null;
      return;
      }
    // here pick is not null, so we have entered something or are still inside
    if (InsideParty == null)
      {
      InsideParty = pick; // we have entered this element
      pick.mouseEntered(e);
      }
    else
      {
      // We were already inside an element
      if(InsideParty != pick)
	{
	// ... but in a new one
	pick.mouseEntered(e);
	InsideParty.mouseExited(e);
	InsideParty = pick;
	return;
	}
      // ... in the same one
      pick.mouseMoved(e);
      }
    }

public void mouseDragged(java.awt.event.MouseEvent e)
    {
      if (DragParty == null)
	{
	  // reserved for future use
	}
      else
	{
	  Point tmp = e.getPoint();
	  Dimension temp = Display.unconvertXY(tmp);
	  // set Event coordinates to grid (for use in AnimElement)
	  e.translatePoint(-tmp.x + temp.width, -tmp.y + temp.height);
	  DragParty.mouseDragged(e);
	}
    }

public void keyTyped(KeyEvent e)
    {
    }

public void keyPressed(KeyEvent e)
    {
    int code = e.getKeyCode();
    if (code == e.VK_BACK_SPACE || code == e.VK_DELETE)
      {
	removeAllSelected();
      }
    }

public void keyReleased(KeyEvent e)
    {
    }


/** This method is called in getAll() and getPeers() etc. to remove all
  elements that have been deleted since the last call. (I.e. the "garbage
  collection" method*/
private void doGC()
    {
    if (!DeleteChange) return; // nothing was deleted
    // getSize done manually - to avoid infinite loop
    Iterator it = Layer.iterator();
    int count = 0;
    while (it.hasNext())
      {
      count += ((Collection) it.next()).size();
      }
    AnimElement[] retval = new AnimElement[count];
    AnimElement curr = null;
    it = Layer.iterator();
    int counter = 0;
    while (it.hasNext())
      {
      Iterator innerit = ((Collection)it.next()).iterator();
      while (innerit.hasNext())
	{
	curr = (AnimElement) innerit.next();
	if ((curr.getState() & AnimElement.DELETED) != 0)
	  {
	  retval[counter] = curr;
	  counter++;
	  }
	}
      }

    for(int i=0; i<counter;i++)
      {
      it = Layer.iterator();  // iterator over the layers
      while (it.hasNext())
	{
	Collection currc = (Collection) it.next(); // get current layer
	currc.remove(retval[i]);
	}
      }
    DeleteChange = false;
    }

/** This method inserts the elements that have been added since
    the last call of doIC().
    The implementation of this is a bit more complicated than
    one might initially assume. To ensure that everything works
    correctly:
    <ul>
      <li> Check if the layer already exists, if not create it.
      <li> If list is already connected to a display, ensure
           that the Animrepresentation is set to the right position
	   and size.
      <li> Set the display variable of the peer.
    </ul> */
protected void doIC()
  {
  if (!InsertChange) return; // acceleration
  Iterator ilist = InsertList.iterator();
  SALNewElement elem;
  AnimElement e;
  int layer;
  while (ilist.hasNext())
    {
    elem  = (SALNewElement) ilist.next();
    e     = elem.Element;
    layer = elem.Layer;
    // Important: check if layer already exists!
    if (layer >= Layer.size())
      {
      // create layers up to the current one
      for(int i= Layer.size(); i <= layer; i++)
	{
	Layer.add(new HashSet());
	}
      }
    // add to layer
    ((Collection)(Layer.get(layer))).add(e);
    // set the return reference
    // -> This has happens already to avoid a bug when doing stuff before first display!


    // set inital scale if possible and repaint to show new component
    if (Display != null)
      {
      e.getPeer().Display = Display;
      e.getPeer().scale(Display.getScale());
      e.getPeer().setPosition(e.getGridPosition());  // 'fire property Change'
      Display.repaint();
      }
    else
      {
      e.getPeer().scale(1.0);
      e.getPeer().setPosition(e.getGridPosition());
      }
    }
  InsertChange = false;
  InsertList.clear();
  }



/** Use this one mainly for debugging purposes */
public String toString()
    {
    return new String("SimpleAnimList with " + getSize() + " elements.");
    }

}


/** This class is used to allow inter-thread communication */
class Communicate
{
private boolean Pause = false;

public synchronized void setPause(boolean p)
    {
    Pause = p;
    if (!p) notify(); // Pause over
    }

public synchronized boolean isPause() throws InterruptedException
  {
  if (Pause) wait();
  return Pause;
  }


}

/** This class is used to store new Elements before they are inserted into
  our actual data structure. This dire step is necessary because of
  concurrency issues, iterators and the unsuccessful cooperation thereof.*/
class SALNewElement
  {
  public AnimElement Element;
  public int Layer;
  }



