
package viror.teach.anim;

import java.awt.Dimension;

/** A helper class that can be used to generate splines. During setup
    the coefficients are generated, computePoint is then used to get
    the points on the curve. All functions are based on the book
    mentioned in computePoint. */
public class Spline
{

  /** The constructor needs the degree of the spline (i.e. number
      of control points minus one) and an array of Controlpoints.
      @param degree of the spline
      @param cPoints are the control points that define the spline */
  public Spline(int degree, Dimension[] cPoints)
    {
    // save degree for future reference
    Degree = degree;
    // compute the coefficients for spline display
    computeCoefficients();
    // set Control Points
    CPoint = cPoints;
    }

  /** This alternative constructor gets us a spline that is a straight
      line. This is a shortcut - any spline that has three points in a row
      will be a line (if degree = 2 ). While drawing lines is more efficient
      than straight splines, this method can be used for motion paths.
      @param start is the beginning of the line
      @param end is the end of the line */
  public Spline(Dimension start, Dimension end)
    {
    CPoint = new Dimension[3];
    CPoint[0] = start;
    CPoint[2] = end;
    // interpolate a third point between the other two
    CPoint[1].height = (start.height + end.height) / 2;
    CPoint[1].width  = (start.width  + end.width)  / 2;
    Degree = 2;
    computeCoefficients();
    }

  /** The standard constructor generates a cubic spline (degree three)
    where all controlpoints are at 0,0. */
  public Spline()
    {
      this(3,new Dimension[4]);
      for(int i=0; i <= 3; i++) CPoint[i] = new Dimension();
    }


  /** The array of control points that define the appearance of a spline. */
  protected Dimension[] CPoint;

  /** Order of the spline (i.e. number of control points minus 1) */
  protected int Degree;

  /** The coefficients of the spline - those are calculated automatically
      within the constructor */
  protected double[] Coefficient;

  /** This internal helper function is used by the constructor to
      calculate the coefficients. */
  protected void computeCoefficients()
    {
    Coefficient = new double[Degree + 1];

    for (int k=0; k <= Degree; k++)
      {
	Coefficient[k] = 1;
	for (int i = Degree; i >= k+1; i--)
	  Coefficient[k] *= i;
	for (int i = Degree - k; i >= 2; i--)
	  Coefficient[k] /= i;
      }
    }

  /** This algorithm does spline calculations and is directly adapted
      from Hearn, Baker (1997) "Computer Graphics - C Version", 2nd ed.
      Prentice Hall, New Jersey (p. 329) It computes a point on a spline
      curve.
      @param u is a value between 0 and 1 which indicates the position
      within the spline curve. A zero is the beginning point of the
      curve, one is the endpoint and anything else are positions in
      between.
      @return The coordinates of the point that has the position u
      within the spline. */
  public Dimension computePoint(double u)
    {
    int n = Degree;
    double blend;
    Dimension retPt = new Dimension(0,0); // the value to be returned

    /* add influence of each control point */
    for(int k=0; k <= Degree; k++)
      {
      // calculate Bernstein polynomial
      blend = Coefficient[k] * Math.pow(u,k) * Math.pow(1.0 - u, n - k);
      // add control point weighed by polynomial to coordinates
      retPt.width += CPoint[k].width * blend;
      retPt.height += CPoint[k].height * blend;
      }
    return retPt;
    }

  /** 'Renders' the curve by returning an array of points. Based on
      resolution, different choices of m seem appropriate. The acutal
      graphical display is up to the user but is simple since one only
      needs a class to plot any points. (Note that it is probably better to
      use java2D functions to actually draw a spline)
      @param anz is the number of points the line should consist of.
      A value of two would only render the endpoints, less than one creates
      undefined behaviour.
      @return The array of points the spline curve consists of. */
  public Dimension[] bezier(int anz)
    {
    int m = anz - 1;
    Dimension[] retPt = new Dimension[anz];

    for (int i=0; i <= m; i++)
      {
      retPt[i] = computePoint((double)i / (double) m);
      }
    return retPt;
    }

  /** returns the degree of this spline */
  public int getDegree()
    {
    return Degree;
    }


  /** Sets the n-th control point to the given value.
   @param n The control point to be set (starts at 0)
   @param value The new value for the point */
  public void setControlPoint(int n, Dimension value) throws
                            ArrayIndexOutOfBoundsException
    {
      if (n > Degree) throw new ArrayIndexOutOfBoundsException();
      CPoint[n] = value;
    }

  /** Get the control point with index n. Warning: since we do
   not return a copy, one should not alter this value, lest
   the spline will be affected */
  public Dimension getControlPoint(int n) throws
                             ArrayIndexOutOfBoundsException
    {
      if (n > Degree) throw new ArrayIndexOutOfBoundsException();
      return CPoint[n];
    }

  /** Returns all control points */
  public Dimension[] getControlPoints()
    {
      return (Dimension[]) CPoint.clone();
    }

  /** use this method for debugging etc. */
  public String toString()
    {
      String ret =  new String("Spline: Control Points - ");
      for (int i = 0; i <= Degree; i++)
	ret = ret.concat("["+CPoint[i].width+"|"+CPoint[i].height+"]");
      ret = ret.concat(" Degree: " + Degree);
      return ret;
    }
}


