
package viror.teach.anim;

import java.awt.*;
import java.awt.geom.*;

/** This class draws a spline on screen. Note that all points beyond
  the first one are given with coordinates relative to it.
  @author Ralf Gitzel, Adriaan Geursen (testActual()) */
public class SplineRep extends VectorRep
{
  /** The start point of the spline in screen coordinates. Note that
   since we position the spline with the AffineTransform Pos from
   VectorRep, this value is always the 0,0 point converted with the
   Display.convertXY method */
  Dimension Start;

  /** the control point values are all relative to the start point in
    grid coordinates. */
  Dimension[] CPoint;

  /** The constructor needs a start point to define the position in
    the display and 3 more control points given as relative coordinates
    to describe the shape of the curve */
  public SplineRep(Dimension start, Dimension[] cpoint)
    {
      CPoint = cpoint;
      updateCurve();
      setPosition(start);
    }

  /** The standard constructor gives us a curve with all points 0,0 */
  public SplineRep()
    {
      CPoint = new Dimension[3];
      for(int i=0; i<3;i++) CPoint[i] = new Dimension();
      updateCurve();
      setPosition(new Dimension());
    }

  /** Overwritten to reset the value of Start */
  public void scale(double scale)
    {
    super.scale(scale);
    if (Display != null) Start = Display.convertXY(new Dimension());
    else Start = new Dimension();
    updateCurve();  // use new start point
    }

  /** Used internally to adapt CubicCurve to new CPoints */
  private void updateCurve()
    {
    /* now compute and set the shape - note that we start at 0,0
       since we move the curve to its position with a transform
       set lateron. */
    int[] x = new int[3];
    int[] y = new int[3];
    for(int i=0; i<3; i++)
      {
      x[i] = CPoint[i].width;
      y[i] = CPoint[i].height;
      }
    Original = new CubicCurve2D.Double(0,0,x[0],
				      y[0], x[1], y[1], x[2], y[2]);
    /* only update BoundingBox if Pos is already defined (the
       case Pos == null only occurs in the constructor.) */
    if (Pos != null) updateBoundingBox();
    }

  /** Sets a control point to another position. The start point
      is set with either setPosition or setStartPoint. */
  public void setControlPoint(int index, Dimension d) throws
    ArrayIndexOutOfBoundsException
    {
    if (index >= 3) throw new ArrayIndexOutOfBoundsException();
    CPoint[index] = d;
    updateCurve();
    }

  /** Sets the startpoint to another position and therefore moves
    the whole spline, since all other points are relative to that one */
  public void setStartPoint(Dimension d)
    {
    setPosition(d);
    }

  /** setPosition ensures that all control points lie correctly
    in relative position to the starting point */
  public void setPosition(Dimension d)
    {
      /* update start point (which is always 0/0! The position is
	stored in Pos, the transformation matrix) */
      if (Display != null) Start = Display.convertXY(new Dimension(0,0));
      else Start = new Dimension(0,0);
      // set up identity matrix
      Pos = new AffineTransform();
      /* multiply transformation matrix. We have to move to
	 the startpoint (out of the no-paint area) + the position
	 adjusted to the screen resolution (via the scale factor) */
      Pos.translate(Start.width + d.width * SFactor,
		    Start.height + d.height * SFactor);
      updateBoundingBox();
    }

  /** takes a spline object and uses it as a new basis for the
    CubicCurve Original */
  public void setToSpline(Spline nu)
    {
      Dimension s = nu.getControlPoint(0);
      for (int i=1; i < 4; i++)
	{
	Dimension d = nu.getControlPoint(i);
	CPoint[i-1].width  = d.width  - s.width;
	CPoint[i-1].height = d.height - s.height;
	}
      updateCurve();
      setPosition(s);
    }

  public boolean testActual(Dimension p)
    {
      //compute grid values for the mouse click
      Point ps      = new Point ((int) Math.round(p.getWidth()), (int) Math.round(p.getHeight()));
      Dimension pg  = Display.unconvertXY(ps);
      //compute grid values for the mouse click
      ps             = new Point ((int) Math.round(Pos.getTranslateX()), (int)
                        Math.round(Pos.getTranslateY()));
      Dimension tran = Display.unconvertXY(ps);
      //from here all values relative to grid (display independent)...
      double clickX = pg.getWidth()  - tran.getWidth();
      double clickY = pg.getHeight() - tran.getHeight();
      //compute length of spline
      double length = 0;
      if (CPoint != null) {
        double cWidth, cHeight;
        for (int i=0; i < (CPoint.length - 1); i++) {
          if ((CPoint[i] != null) && (CPoint[i+1] != null)) {
            cWidth  = Math.abs(CPoint[i].getWidth()  -
              CPoint[i+1].getWidth());
            cHeight = Math.abs(CPoint[i].getHeight() -
              CPoint[i+1].getHeight());
            length += Math.sqrt(cWidth*cWidth + cHeight*cHeight);
          }
        }
      }
      //length = 0 -> nothing to select...
      else return false;

      //the spline will be selected, even if it is epsilon pixels away from the line
      double epsilon = 3.0;

      //check if all points are set correctly
      int defNr;
      for (defNr = 0; defNr < CPoint.length; defNr++) {
        if (CPoint[defNr] == null) break;
      }
      defNr--;
      Dimension [] tCPoint = new Dimension[defNr+2];
      tCPoint[0] = new Dimension();
      for (int i=0; i <= defNr; i++) tCPoint[i+1] = CPoint[i];
      Spline actSpline = new Spline(defNr, tCPoint);

      //run through the spline to check, if line runs through the epsilon-box around the pixel p
      Dimension cp;
      for (int i=0; i < length; i++) {

        cp = actSpline.computePoint(i/length);
        //System.out.println(cp.getWidth() + " " + cp.getHeight());

        if ((cp.getWidth()  < clickX + epsilon) && (cp.getWidth()  >
            clickX - epsilon)
         && (cp.getHeight() < clickY + epsilon) && (cp.getHeight() >
           clickY - epsilon))
             return true;

      }
      return false;
    }


  /** Implementation Notes: Updating the bounding box is easy in
    this case thanks to the getBounds2D method. Note that having
    negative spline points will require special attention -
    see source for details. */
  protected void updateBoundingBox()
    {
      // do basic algorithm
      double x = Pos.getTranslateX();
      double y = Pos.getTranslateY();
      Rectangle2D temp = Original.getBounds2D();
      double w = temp.getWidth()  * SFactor;
      double h = temp.getHeight() * SFactor;
      /* Special case: some spline points lie in negative regions,
	 either x or y wise -> the axis that has negative values
	 will have to be multiplied by -1 (the outer point) */
      boolean doX = false, doY = false;
      for (int i = 0; i < 3; i++)
	if (CPoint[i].width < 0) doX = true;
      for (int i = 0; i < 3; i++)
	if (CPoint[i].height < 0) doY = true;
      // decrease x coordinate
      if (doX) x -= w;
      // decrease y coordinate
      if (doY) y -= h;

      BoundingBox = new Rectangle2D.Double(x,y,w,h);
    }


  /** Use this method for debugging etc. */
  public String toString()
    {
      String ret = new String("SplineRep: Pos(");
      ret = ret.concat(Start.width+"|"+Start.height+") ");
      for (int i = 0; i < 3; i++)
	ret = ret.concat("["+CPoint[i].width+"|"+CPoint[i].height+"]");
      return ret;

    }

  /** Create a copy of this SplineRep but make sure that shape is
    updated and array is cloned */
  public Object clone()
    {
    SplineRep cpy = null;
    try {
      cpy = (SplineRep) super.clone();
      cpy.CPoint = new Dimension[CPoint.length];
      for (int i=0; i < CPoint.length; i++)
	cpy.CPoint[i] = new Dimension(CPoint[i].width, CPoint[i].height);
      cpy.updateCurve();
    } catch (Exception e)
      {
      System.err.println("SplineRep.clone failed: "+e);
      }
    return cpy;
    }


}

