
package viror.teach.anim;

import java.awt.geom.*;
import java.awt.*;

/** This class encompasses all non-bitmap based forms of
    AnimRepresentations. It is possible to set the line
    style and the fill style of the objects here, either
    directly using java2d classes or via various helper
    methods. VectorRep has a couple of factory methods
    that will create vastly different results. Users
    are not to directly create objects of the subclasses
    other than Spline. Note that all Java2D is hidden
    completey from the user as in all other cases since
    it is somewhat confusing to the neophyte.
    @Author: Ralf Gitzel, Adriaan Geursen (testActual()-Method) */
public abstract class VectorRep extends AnimRepresentation
{
  /** This primitive flag indicates a square shape */
  public static final int SQUARE = 1 ;

  /** This primitive flag indicates a circular shape */
  public static final int CIRCLE = 2 ;

  /** AND this flag to the original one to fill element */
  public static final int FILL = 4;

  /** The element to be drawn here. Note that position is
   0,0 to ensure proper scaling and that positioning will
   be done via a translation. */
  protected Shape Original;

  /** The flood fill properties of this element */
  protected Paint Filling = Color.black;

  /** The highlight colour value */
  protected Paint HighlightColour = Color.red;

  /** The stroke used by the objects */
  protected Stroke Pencil;

  /** If this flag is true, then the object is produced if
    g.fill else, g.draw is used */
  protected boolean DoFill;

  /** The scale factor of this element saved in matrix form */
  protected AffineTransform Scale;

  /** The position of this element is saved as a translation
    matrix */
  protected AffineTransform Pos;

  public void paint(Graphics2D g)
    {
      if (Cloaked) return;
      // save old stroke and old paint
      Paint  oldP = g.getPaint();
      Stroke oldS = g.getStroke();
      // set our stroke and paint if defined
      if (Pencil  != null) g.setStroke(Pencil);
      if (Highlighted) g.setPaint(HighlightColour);
      else if (Filling != null) g.setPaint(Filling);
      // transform matrices have opposite order...
      AffineTransform mat = new AffineTransform(); // identity Matrix
      mat.concatenate(Pos);    // moved to proper position
      mat.concatenate(Scale);  // scaling multiplied
      g.setTransform(mat);     // apply matrix
      if (DoFill) g.fill(Original);
      else g.draw(Original);
      g.setTransform(new AffineTransform()); // set back to identity
      g.setPaint(oldP);
      g.setStroke(oldS);
    }


  /** this factory method generates a geometric primitive.
    @param type The type can be any 'primitive' flag (such
    as SQUARE, CIRCLE)
    @param size The length of the sides of this primitive */
  public static VectorRep createPrimitive(int type, int size)
    {
    return new PrimitiveRep(type, size);
    }

  /** This factory method generates a spline */
  public static VectorRep createSpline(Dimension start, Dimension[] cpoint)
    {
    return new SplineRep(start, cpoint);
    }

  /** This factory method generates a polygon, which can be filled or not */
  public static VectorRep createPolygon(Dimension[] points, boolean fill)
    {
      return new PolygonRep(points, fill);
    }

  /** This implementation of scale uses AffineTransforms.*/
  public void scale(double scale)
    {
    Scale = new AffineTransform();
    Scale.scale(scale,scale);
    SFactor = scale;
    updateBoundingBox();
    }

  /** allows to set the fill style. For non java2D experts it should
    be noted that the Color class is derieved from Paint and thus
    this method can be used to set colour. */
  public void setPaint(Paint nuPaint)
    {
      Filling = nuPaint;
    }

  public Paint getPaint()
    {
      return Filling;
    }

  /** set the stroke style (a java2D feature) */
  public void setStroke(Stroke s)
    {
    Pencil = s;
    }

  /** returns the current stroke style */
  public Stroke getStroke()
    {
    return Pencil;
    }

  /** A shortcut function to set line width. Note that if a selfmade
   stroke was setted this function will overwrite the custom settings; */
  public void setLineWidth(double width)
    {
    Pencil = new BasicStroke((float)width);
    }

  /** Sets the colour that will be used to highlight the component.
      Default colour is red. */
  public void setHighlightColor(Color c)
    {
      HighlightColour = c;
    }

  /** the standard way of positioning an element (such that the coordinates
    are the object center). Some derieved classes such as SplineRep
    will overwrite this one, so the assumption that gridposition is
    center of the object is not automatically true. */
  public void setPosition(Dimension d)
    {
    Dimension d2 = null;
    if (Display != null) d2 = Display.convertXY(d);
    else d2 = (Dimension) d.clone();
    // remember - position gives center but
    // drawing has to be in the corner
    Rectangle2D b = Original.getBounds2D();
    double wide = SFactor * b.getWidth() / 2.0;
    double hi   = SFactor * b.getHeight() / 2.0;
    double x = d2.width  -  wide;
    double y = d2.height -  hi;
    Pos = new AffineTransform();
    Pos.translate(x,y);
    updateBoundingBox();
    }



  /** Implementation Notes: Updating the bounding box is easy in
    this case thanks to the getBounds2D method.
    SplineRep overwrites this method since it requires special
    consideration. */
  protected void updateBoundingBox()
    {
      // do basic algorithm
      double x = Pos.getTranslateX();
      double y = Pos.getTranslateY();
      Rectangle2D temp = Original.getBounds2D();
      double w = temp.getWidth()  * SFactor;
      double h = temp.getHeight() * SFactor;

      BoundingBox = new Rectangle2D.Double(x,y,w,h);
    }

  /** By default Vectors use their bounding box to check for
    intersection with a point - derieved classes are expected to
    overwrite this one! */
  public boolean testActual(Dimension d)
    {
    return testBoundingBox(d);
    // *****************************************************
    /* Attention: in the real implementation do not forget to
       multiply the inverted matrices to d !!!! */
    // *****************************************************
    }

}


/** The representation class for geometric primitives. Sanity Saving
 Note: If you look for the source code, it is in VectorRep.java
 (the PrimitiveRep class is not public) - one would not really need
 a new class for this but the things normally done in the
 factorymethod have been put here for clarity. */
class PrimitiveRep extends VectorRep
{

  /** Do not call the constructor directly but use the factory
    methods of VectorRep instead */
  protected PrimitiveRep(int type, int size)
    {
    /* we can set position to zero since we move the object via
       a transform */
    int swit = type & 3;
    switch(swit) // get only first two bits
      {
      case SQUARE:
	Original = new Rectangle2D.Float(0,0,size, size);
        break;
      case CIRCLE:
	Original = new Ellipse2D.Float(0,0, size, size);
	break;
      default:
	System.err.println("False definition!");
      }
    /* check for fill */
    if ((type & FILL) == FILL) DoFill = true;
    else DoFill = false;
    /* set other desired properties to acceptable default values */
    Pencil  = new BasicStroke(1.0f);
    Filling = Color.black;
    /* this avoids NullPointerExceptions when setPosition or scale
       are called, since both will update BoundingBox and the
       BBox calculation uses both values. */
    Pos = new AffineTransform();
    Scale = new AffineTransform();
    }

public boolean testActual(Dimension d)
    {
      //test if this is a circle or not
      if (Original instanceof java.awt.geom.Ellipse2D.Float)
      {
        //Original.getBounds2D().getX()
        double posX    = this.BoundingBox.getX();
        double posY    = this.BoundingBox.getY();
        double width   = this.BoundingBox.getWidth();
        double height  = this.BoundingBox.getHeight();
        double centerX = posX + width  / 2;
        double centerY = posY + height / 2;
        double clickX  = d.getWidth()  - centerX;
        double clickY  = d.getHeight() - centerY;
        double radius;
        //test if round circle or elipse
    if (Math.abs(height - width) < 0.00001 ) radius = height/2;
        else
        {
          double alpha = Math.atan(clickY / clickX);
                 // (h/2*sin(a))² + (w/2*cos(a))²
                 radius = (height*height/4 *
                    Math.sin(alpha)*Math.sin(alpha)) + (width*width/4 *
                    Math.cos(alpha)*Math.cos(alpha));
        }
        //test if pixel is inside circle...
        if (radius < Math.sqrt(clickX*clickX + clickY*clickY))
        {
          return false;
        }
        else
        {
          return true;
        }
      }
      else return testBoundingBox(d);
    }

}



