

package viror.teach.anim.net;

import viror.teach.anim.*;
import java.awt.Image;
import java.util.*;

/**
 * The basic class for all routers etc. The most important method here is
 * handlePackage(package p). The primary reason for this class is to give
 * AbstractLink a reference point.
 * @author Ralf Gitzel
 * @version 1.0
 */
public abstract class AbstractRouter extends AnimStatic implements Progressive
{

private String Name;

private String Description;

/** The List of all events that can occur. While this is probably not the smartest
 *  implementation for this (i.e. using a list) it should do for the basic version. */
private List Events;

/** The number of turns until the next event is executed */
private int NextEvent;

/** The number of turns until the next turn is executed. Unlike NextEvent this
 *  value does not decrease, i.e. in the beginning NextSaved and NextEvent are
 *  the same; NextSaved stores the original counter value of NextEvent. */
private int NextSaved;

/** The set of all timer listeners */
private Set TimerListeners;

/** The set of all ProgressIndicators */
private Set ProgressIndicators;

/** The table of all progress reactions */
private Map ProgressTable;

/** A progress table flag */
public static int FORWARD = 0;

/** A progress table flag */
public static int BACK = 1;

/** A progress table flag */
public static int ABSOLUTE = 2;

/** The default entry to use when no match in table */
TableEntry DefaultTableEntry = null;

/** The last package that has arrived so far. */
private Package LastNewPackage;

public AbstractRouter(Image normal, Image selected)
  {
  this(normal, selected, true,  0, 0);
  }

public AbstractRouter(Image normal, Image selected, boolean dragable, int x, int y)
  {
  super(normal, selected, dragable, x, y);
  Events = new LinkedList();
  NextEvent = Integer.MAX_VALUE;
  TimerListeners = new HashSet();
  ProgressIndicators = new HashSet();
  ProgressTable = new HashMap();
  }

/** What should this router do when a package arrives? The implementation in this
 *  class only sets the LastNewPackage to p. If it is overwritten, super.packageArrived()
 *  should be called, to notify propertyChangeListeners of the new package. Also the
 *  progress table is evaluated here. */
public void packageArrived(Package p)
  {
  TableEntry t = progressTableLookup(p.getType());
  if (t != null)
    {
    if (t.Forward == ABSOLUTE) this.fireProgress(t.Channel, t.Steps);
    else for (int i = 0; i < t.Steps; i++)
          {
           if (t.Forward == FORWARD) fireProgress(t.Channel, true);
           if (t.Forward == BACK)    fireProgress(t.Channel, false);
          }
    }
  setLastNewPackage(p);
  }

/** The name of this element */
public String getName()
  {
  return Name;
  }

public void setName(String name)
  {
  Name = name;
  }

/** The description of this element */
public String getDescription()
   {
   return Description;
   }

public void setDescription(String txt)
  {
  Description = txt;
  }

/** Add a new timer event to the list. Timer event in the current version can
 *  send packages only - future relases might change this.
 *  @param pack The package to be sent
 *  @param code An arbitrary int value that may be used to describe situations to
 *              the listener.
 *  @param time The number of animation cycles before this event should happen. The
 *              value Integer.MAX_VALUE is explicitly disallowed for time!
 */
public void addTimerEvent(Package pack, int code, int time)
  {
  TimerEvent t = new TimerEvent();
  t.Send = pack;
  t.CodeValue = code;
  t.Time = time;
  Events.add(t);
  if (NextEvent > time)
    {
    NextEvent = time;
    NextSaved = time;
    }
  }

/** sets the package that should be considered the last one that has arrived
 *  so far. This method is called by handleArrival(). */
protected void setLastNewPackage(Package p)
  {
  propertyChangeListeners.firePropertyChange("LastNewPackage", LastNewPackage, p);
  LastNewPackage = p;
  }

public Package getLastNewPackage()
  {
  return LastNewPackage;
  }
/** The periodic update of the physical representation. In this case nothing
 *  happens...
 *  Also timer events are handled here -> all due events fire a timer event
 *  that timer listeners can handle. */
public void move()
  {
  if (NextEvent == Integer.MAX_VALUE) return; // event queue empty
  if (NextEvent == 0)
    {
    // act out all events listed here...
    Set GoodBye = new HashSet(); // store all elements to be deleted here.
    int neuTimer = 0;  // the new value of NextEvent after this cycle
    Iterator it = Events.iterator();
    TimerEvent t = null;
    // decrease all timers and execute events that are due
    while (it.hasNext())
      {
      t = (TimerEvent) it.next();
      t.Time -= NextSaved;
      if (t.Time <= 0)
        {
        fireTimerEvent(t);
        t.timeIsUp();
        GoodBye.add(t);
        }
      else if (t.Time > neuTimer) neuTimer = t.Time; // might be due the next time...
      }
    // Delete those that are in the GoodBye list (i.e. have been due)
    Iterator del = GoodBye.iterator();
    while(del.hasNext())
      {
      Events.remove(del.next());
      }
    // check if queue is empty
    if (Events.size() == 0)
      {
      NextEvent = Integer.MAX_VALUE;
      NextSaved = 0;
      }
    else
      {
      NextEvent = neuTimer;
      NextSaved = neuTimer;
      }
    }
  NextEvent--; // prepare for next cycle
  }

/** Sends a message to all TimerListeners registered here. */
protected void fireTimerEvent(TimerEvent t)
  {
  Iterator it = TimerListeners.iterator();
  TimerListener tl = null;
  while (it.hasNext())
    {
    tl = (TimerListener) it.next();
    tl.timerEventOccured(t);
    }
  }

public void addTimerListener(TimerListener t)
  {
  TimerListeners.add(t);
  }

public boolean removeTimerListener(TimerListener t)
  {
  return TimerListeners.remove(t);
  }

public void addProgressIndicator(ProgressIndicator i)
  {
  ProgressIndicators.add(i);
  }

  public void removeProgressIndicator(ProgressIndicator i)
  {
  ProgressIndicators.remove(i);
  }

public void fireProgress(int stateAffected, boolean forward)
  {
  Iterator it = ProgressIndicators.iterator();
  while (it.hasNext())
    {
    if (forward) ((ProgressIndicator) it.next()).advance(stateAffected);
    else ((ProgressIndicator) it.next()).stepBack(stateAffected);
    }
  }

public void fireProgress(int stateAffected, int line)
  {
  Iterator it = ProgressIndicators.iterator();
  while (it.hasNext())
    {
    ((ProgressIndicator) it.next()).setPosition(stateAffected, line);
    }
  }

/** The AbstractRouter has an internal table that is sorted by package types.
 *  Whenever a package arrives, the table is checked to see whether said package
 *  should fire a progress. The nature of this is specified in detail in the table
 *  entry, whose values are set here.
 *  @param pack The package type which causes the following actions when it arrives
 *  @param channel The progress channel to be altered by this arrival.
 *  @param forward If FORWARD, the channel will be moved forward, if BACK it will step
 *                 back. ABSOLUTE means that it will jump to the given position (which
 *                 is the same as resetting to 0 and then moving steps forward.
 *  @param steps The number of steps to be taken in the specified direction.
 */
public void addProgressTableEntry(int pack, int channel, int forward, int steps)
  {
  ProgressTable.put(new Integer(pack), new TableEntry(channel, forward, steps));
  }

/** The default progress action to be taken when a package arrives that is not
 *  specified in the progress table. If none is set, no default action will occur. */
public void addProgressTableDefaultEntry(int channel, int forward, int steps)
  {
  DefaultTableEntry = new TableEntry(channel, forward, steps);
  }

/** removes the default table entry. */
public void removeProgressTableDefaultEntry()
  {
  DefaultTableEntry = null;
  }

/** A lookup in the ProgressTable - will return a default if it has been defined,
 *  might return null otherwise. */
protected TableEntry progressTableLookup(int ptype)
  {
  TableEntry t = (TableEntry) ProgressTable.get(new Integer(ptype));
  if (t == null) t = DefaultTableEntry;
  return t;
  }


/////////////////////////// INNER CLASSES /////////////////////////////

/** This inner class is used to describe timer events that can occur within
 *  routers. The time segments used here are simulation time based, i.e they
 *  are not given in seconds but rather in animation cycles. When the event
 *  occurs, the method timeIsUp() is called which - in this version - is emtpy.
 *  Subclasses may do fancy things here however, such as sending packages etc. */
public class TimerEvent
  {
  /** A package can be associated with a timer event - usually to be sent, but other
   *  uses are possible, since the base class does not handle the timer events directly */
  public Package Send;

  /** The time from the creation until now to execution. Will be zero by the time
   *  the event is due but will NOT be updated every turn. */
  public int Time;

  /** A code value for user defined actions. */
  public int CodeValue;

  /** This is executed when a timer event is due. Currently there is not much use for
   *  it, but subclasses might overwrite this. */
  public void timeIsUp()
    {
    }
  }

/** This inner class is used to describe entries in the progress table. The progress
 *  table looks like this
 *  <pre>
 *  Package Type (Map key)  |  Channel | Forward? | Steps
 *  --------------------------------------------------------
 *         int              |   int    | int      | int
 *  </pre>
 *  The values are explained below. The idea is that when a package arrives there is
 *  a lookup in the table to see how to react in terms of progress. It is also possible
 *  to react with a progress to a timer event but not through this table. Rather one
 *  has to create a TimerListener that will fire a progress manually. */
protected class TableEntry
  {
  /** The channel to be used in the progress listener. See ProgressListener interface
   *  for details. */
  public int Channel;

  /** This value decides about the progress direction. Possibilities are
   *  FORWARD, BACK and ABSOLUTE which are explained in addProgressTableEntry */
  public int Forward;

  /** The number of steps of progress to be taken in the stated direction */
  public int Steps;

  public TableEntry(int channel, int forward, int steps)
    {
    Channel = channel;
    Forward = forward;
    Steps   = steps;
    }
  }


}