
package viror.teach.anim.net;

import viror.teach.anim.*;
import javax.swing.JComponent;
import java.beans.*;


/** A connection between two routers. This may be a direct link
 * or a whole subnet, a high-level protocol or a very basic bit-level
 * connection. The connection has different attributes that are described
 * below. Note that ErrorRate is not used directly in any way but rather
 * should be used by specific packages - i.e. the package uses getErrorRate
 * and reacts according to the value by flipping bits etc. Cost is not used
 * neither but might come in handy to test routing protocols etc. */
public class Link extends AbstractLink
{


/** The bandwidth and the size of the package determine how long
  it will take before a package that is on the link starts
  moving (i.e. the time until the last byte is on the link).
  Waiting time is size / bandwidth (integer division) */
private int Bandwidth;

/** The cost associated with this link. packages ignore cost
  but classes that are derieved from package might read this
  information and use it (e.g. increase their hopcounter). */
private int Cost;

/** The loss rate - the percentage chance that packages might
  be lost along the way. */
private double LossRate;

/** The chance for bit errors etc. Bit errors are not introduced
  by the link in placePackage, rather the packages themselves have
  to decide how to do it in assessLink. */
private double ErrorRate;

  ///////////////////////////////

/** The standard link constructor. The link will be straight and
  represented by a spline. */
public Link()
    {
    super();
    generateRepresentation();
    Delay = 100;
    Bandwidth = Integer.MAX_VALUE;
    Cost = 0;
    LossRate = 0.0;
    ErrorRate = 0.0;
    addAllowedStartPoint("Router");
    addAllowedEndPoint("Router");
    setIdentification("Link");
    }

/** Sends a package p from s to the other end of the link.
  @param s The end of the link where the message originates.
           All other routers will have no effect!
  @param p The package to be send. */
public void placePackage(Router s, Package p)
    {
    p.setWaitTime(p.getPSize() / Bandwidth);
    // set up loss
    if (Math.random() <= LossRate) p.willBeLost();
    // let package do its thing
    p.assessLink(this);
    // the standard placePackage part
    super.placePackage(s, p);
    }


  ////////// sets and gets //////////////

/** Sets the bandwidth value, i.e. how many units of package
  size the link can have written on per animation cycle */
public void setBandwidth(int b) throws PropertyVetoException
    {
    if (b < 0)
      throw new PropertyVetoException("Positive values only!",
	new PropertyChangeEvent(this, "Bandwidth", new Integer(Bandwidth),
				new Integer(b)));
    Bandwidth = b;
    }

/** Gets the bandwidth value, i.e. how many units of package
  size the link can have written on per animation cycle */
public int getBandwidth()
    {
    return Bandwidth;
    }

/** Sets the loss rate, i.e. what chance a package has to be lost along
  the way. */
public void setLossRate(double e) throws PropertyVetoException
    {
    if (e < 0.0 || e >= 1.0)
      throw new PropertyVetoException("0.0 to 1.0 only!",
	new PropertyChangeEvent(this, "LossRate", new Double(ErrorRate),
				new Double(e)));
    LossRate = e;
    }


/** Gets the loss rate, i.e. what chance a package has to be lost along
  the way. */
public double getLossRate()
    {
    return LossRate;
    }

/** Sets the Cost of this link. This value is not used directly, rather
  packages can use it for whatever they like (i.e. HopCounter, Pruning..) */
public void setCost(int c)
    {
    Cost = c;
    }

/** Gets the Cost of this link. This value is not used directly, rather
  packages can use it for whatever they like (i.e. HopCounter, Pruning..) */
public int getCost()
    {
    return Cost;
    }


/** Sets the error rate of this link. This value is not used but packages
  may read it and react by flipping bits etc. */
public void setErrorRate(double e) throws PropertyVetoException
    {
    if (e < 0.0 || e >= 1.0)
      throw new PropertyVetoException("0.0 to 1.0 only!",
	new PropertyChangeEvent(this, "ErrorRate", new Double(ErrorRate),
				new Double(e)));
    ErrorRate = e;
    }

/** Gets the error rate of this link. This value is not used but packages
  may read it and react by flipping bits etc. */
public double getErrorRate()
    {
    return ErrorRate;
    }

public JComponent getCustomizer(boolean i)
    {
    LinkCustomizer l =  new LinkCustomizer(i);
    l.setObject(this);
    return l;
    }

public String toString()
  {
  return "Link("+getStart()+","+getEnd()+")";
  }

}
