

package viror.teach.anim.net;

import viror.teach.anim.*;
import java.awt.*;
import javax.swing.*;
import java.util.*;
import gitzel.util.*;
import java.net.*;
import java.io.IOException;
import java.beans.PropertyVetoException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;


/**
 * This class represents an object oriented software component, i.e. either
 * a class or an object according to interpretation. The expansion to the
 * standard SoftwareComponent is the ability to call methods via buttons
 * A 'method' is not really a function but an entry in the MethodMap table,
 * that maps method names to packages (which conform to the norms set for
 * SoftwareHandlers.
 * A SoftwareComponent can only be linked to one Router. Trying to link it
 * to another one will cause an IndexOutOfBoundsException.
 * @author Ralf Gitzel
 * @version 1.0
 */
public class OOComponent extends SoftwareComponent implements ActionListener
{

/** The table of methods */
private Map MethodTable = new HashMap();

/** The table of icons to be used with the method on the buttons. In the current
 *  version the package images are used */
private Map MethodIcons = new HashMap();

/** The table of properties */
private Map PropertyTable = new HashMap();

/** If this is set to true, the buttons in the OOComponentCustomizer will be
 *  enabled, otherwise they will not be active. */
private boolean MethodsEnabled = true;

/** If this is true, packages can only be sent synchronized, i.e. while a package
 *  is on the way, the buttons in the OOComponentCustomizer will be set to grey. */
private boolean Synchronized = false;

/** The customizer used with this object */
private OOComponentCustomizer MrCustom;

/** This is the customizer used in a beans context (editable = true) */
private OOComponentCustomizer EditCus;

////////////////////////////////////

/** This constructor is used if the position of the component is not
 *  important yet. */
public OOComponent(Image basic, Image selected)
  {
  this(basic, selected, 0, 0);
  }

/** This is the constructor most likely to be used in manual coding. */
public OOComponent(Image basic, Image selected, int x, int y)
  {
  super(basic, selected, x, y);
  MrCustom = new OOComponentCustomizer(false);
  MrCustom.setObject(this);
  addPropertyChangeListener(MrCustom);
  EditCus = new OOComponentCustomizer(true);
  EditCus.setObject(this);
  addTimerListener(this);
  }

///////////////////////////


/** Please Note: When the customizer is not editable it sports a button panel.
 *  Also note that only one customizer exists - there is no new instanciation
 *  every time the method is called as is usually the case. The reason is that
 *  otherwise (due to the listener relationship) the customizers would never
 *  be deleted. */
public JComponent getCustomizer(boolean editable)
  {
  if (editable)
    {
    return EditCus;
    }
  return MrCustom;
  }




/** Adds a new entry to the method table. Note: the PropertyChange that is fired
 *  here has no old value (it is null). This is pure laziness
 *  but since this is very unlikely to be listened to by anything else than
 *  OOComponentCustomizer it did not seem worth the hassle.
 *  @param name The name under which the method is identified.
 *  @param package The package that should be sent when this method is called.
 *                 Note that if another package was defined beforehand, it will
 *                 not be overwritten, rather both will be sent. Use removePackage
 *                 to cancel certain sending orders. */
public void addMethod(String name, Package p)
  {
  Package[] old = (Package[]) MethodTable.get(name);
  Package[] neu = null;
  if (old == null)
    {
    neu = new Package[1];
    neu[0] = p;
    }
  else
    {
    neu = new Package[old.length+1];
    neu[old.length] = p;
    for (int i=0; i < old.length; i++)
      neu[i] = old[i];
    }
  MethodTable.put(name, neu);
  this.propertyChangeListeners.firePropertyChange("MethodTable", null, MethodTable);
  /* In this case the signal has to be fired after the change since otherwise
     MethodTable still sports the old values */
  }

/** This will remove a method and all the packages it should have sent! Ture is
 *  returned if the operation was sucessful. */
public boolean removeMethod(String name)
  {
  Object tmp = MethodTable.remove(name);
  if (tmp != null) return true;
  return false;
  }

/** Returns the package that should be sent when the given method is "called" */
public Package[] getPackagesForMethod(String methodName)
  {
  Package[] read =  (Package[]) MethodTable.get(methodName);
  Package[] ret  = new Package[read.length];
  for (int i = 0; i < read.length; i++)
    {
    ret[i] = (Package) read[i].clone();
    }
  return ret;
  }

/** Sets an image for a certain method - will be used in the Customizer. */
public void addMethodIcon(String name, Icon x)
  {
  MethodIcons.put(name, x);
  }

/** Return the Icon reserved for a certain method */
public Icon getMethodIcon(String name)
  {
  return (Icon) MethodIcons.get(name);
  }

/** If this is true, packages can only be sent synchronized, i.e. while a package
 *  is on the way, the buttons in the OOComponentCustomizer will be set to grey. */
public void setSynchronized(boolean synch)
  {
  Synchronized = synch;
  }

/** If this is true, packages can only be sent synchronized, i.e. while a package
 *  is on the way, the buttons in the OOComponentCustomizer will be set to grey. */
public boolean isSynchronized()
  {
  return Synchronized;
  }

/** Adds a new property to the table of properties. See the propertyChange-note
 *  in addMethod for oddities */
public void addProperty(String name, String value)
  {
  PropertyTable.put(name, value);
  this.propertyChangeListeners.firePropertyChange("Property", null, PropertyTable);
  }

/** Remove a property from the table of properties. */
public void removeProperty(String name)
  {
  PropertyTable.remove(name);
  this.propertyChangeListeners.firePropertyChange("Property", null, PropertyTable);
  }

public String getPropertyValue(String name)
  {
  return (String) PropertyTable.get(name);
  }

/** Returns all 'method' names as an array */
public String[] getMethodNames()
  {
  Set s = MethodTable.keySet();
  Iterator it = s.iterator();
  String[] retval = new String[s.size()];
  int i=0;
  while (it.hasNext())
    {
    retval[i] = (String) it.next();
    i++;
    }
  return retval;
  }

/** Returns all 'property' names in an array */
public String[] getPropertyNames()
  {
  Set s = PropertyTable.keySet();
  Iterator it = s.iterator();
  String[] retval = new String[s.size()];
  int i=0;
  while (it.hasNext())
    {
    retval[i] = (String) it.next();
    i++;
    }
  return retval;
  }

/** If this is set to true, the buttons in the OOComponentCustomizer will be
 *  enabled, otherwise they will not be active. */
public void setMethodsEnabled(boolean meth)
  {
  this.propertyChangeListeners.firePropertyChange("MethodsEnabled", MethodsEnabled, meth);
  MethodsEnabled = meth;
  }

/** If this is set to true, the buttons in the OOComponentCustomizer will be
 *  enabled, otherwise they will not be active. (Excuse the terrible grammar -
 *  consistency reasons) */
public boolean isMethodsEnabled()
  {
  return MethodsEnabled;
  }




/** This method parses the info for this OOComponent from an URL. The syntax
 *  for the text file should be the following:
 *  M1: first method name
 *  P1: package type to be sent for first method, second package type, ....
 *  I1: the name of the image (as stored in the networksemantics to use) -> if undefined no pic is used, second image...
 *  T1: name of the target component, second target
 *  M2: ....
 *  ....
 *  Pr1: the first property
 *  Pv1: value of the first property
 *
 *  While this may seem tedious, it is human-readable and can be altered without
 *  recompiling..
 *  Also note that a method may have several entries per value - if it is to send several
 *  packages at once. The will be sent in the order given here with a slight time
 *  delay. (To ensure that the do not overlap if on the same line)  */
public void initFromURL(URL data) throws IOException, NumberFormatException,
                                         PropertyVetoException
  {
  IniReader i = new IniReader(data.openStream());
  // set all Methods...
  String name = "";
  String targetName = "";
  String image = "";
  double scale = 0;
  Package p = null;
  Image tmp = null;
  IDPayload target = null;
  IDPayload sender = new IDPayload(getName());
  int ptype = 0;
  int counter = 1;
  while(true)
    {
    name = i.getEntry("M"+counter);
    if (name == null) break;
    StringTokenizer tt = new StringTokenizer(i.getEntry("T"+counter));
    StringTokenizer tp = new StringTokenizer(i.getEntry("P"+counter));
    StringTokenizer ti = new StringTokenizer(i.getEntry("I"+counter));
    if (tt.countTokens() != tp.countTokens() || tp.countTokens() != ti.countTokens())
      {
      System.err.println(this+".initFromURL -> Error in method description "+counter);
      System.err.println("    Number of packages not consistent in T P and I!");
      break;
      }
    while (tt.hasMoreElements())
      {
      targetName = tt.nextToken();
      target = new IDPayload(targetName);
      ptype = new Integer(tp.nextToken()).intValue();
      image = ti.nextToken();
      if (image != null)
        {
        tmp = NetworkSemantics.getSemantics().getImage(image);
        scale = NetworkSemantics.getSemantics().getScale(image);
        if (tmp == null)
          {
          System.err.println(this + ".initFromURL(): "+image+ "is not stored in NetworkSemantics!");
          return;
          }
        p = new Package(ptype, 0, tmp, tmp);
        p.scalePeer(scale);
        }
      else p = new Package(ptype, 0, Color.blue);
      p.addPayload(target);
      p.addPayload(sender);
      this.addMethod(name, p);
      // also update icon list
      this.addMethodIcon(name, new ImageIcon(tmp));
      }
    counter++;
    }
  }


/** The ActionListener is attached to the Customizer and will react to the buttons
 *  and launch the corresponding method call - i.e. package */
public void actionPerformed(ActionEvent e)
  {

  if (e.getSource() instanceof JButton)
    {
    System.out.println("Button pressed!");
    Package[] toSend = getPackagesForMethod(((JButton) e.getSource()).getText());
    if (toSend == null)
      {
      System.err.println(this + " has no package for button " + ((JButton) e.getSource()).getText());
      return;
      }
    if (Synchronized) MrCustom.setEnabled(false);
    // queue these into the send list...
    for (int i=0; i < toSend.length; i++)
      {
      this.addTimerEvent(toSend[i], 0, (i+1) * 3);
      }
    }

  if (e.getSource() instanceof JTextField)
    {
    String[] prop = getPropertyNames();
    for (int i=0; i< prop.length; i++)
      this.addProperty(prop[i], MrCustom.getPropertyField(prop[i]));
    }
  }

public void packageArrived(Package p)
  {
  super.packageArrived(p);
  if (Synchronized && !MrCustom.isEnabled()) MrCustom.setEnabled(true);
  }


}