
package viror.teach.anim.net;

import viror.teach.anim.*;
import java.beans.*;
import java.awt.*;
import javax.swing.*;
import java.util.*;

/** Packages are represented by AnimActors that have additional
  information attached. The package can be given a graphic otherwise
  it will be a square of the chosen colour. Type and payload interpretation
  depend on the application using the library. Note that changing the
  values is only possible when package is still at completion zero
  or at completion = traveltime, i.e. the package cannot be changed
  while travelling, only at the beginning or at the end. */
public class Package extends AnimActor
{

/** The type of the package. This is application dependent, i.e. which
    number requires which interpretation is up to the user. This value
    is used intensively by the SimpleParseHandler class */
private int Type;

/** The payload can be empty but can be used by applications to simulate
    package content. It is an array of objects. */
private java.util.List Payload;

/** The package size. This value is used by links to determine bandwidth
    usage or can be used to determine time lag in the ComplexRouter class. */
private int PSize;

/** The package will disappear when it reaches the described completion value,
 * i.e. it is lost due to faulty data transmission. The default value is -1, a
 * value that is never reached -> The package will not be lost in the default
 * setting. To cause random position package loss, call willBeLost()
 */
private int LossPoint = -1;

////////// constructors ////////////

/** The standard constructor will set default values. This one is
  used primarily with bean-tools. */


/** This constructor is the one to usually use. It will create a
  bland looking square as a package.
  @param type The package type
  @param size The PSize of the package, <i>not</i> the physical size in
  the animation
  @param col The colour of the package*/
public Package(int type, int size, Color col) throws PropertyVetoException
    {
    super(10, VectorRep.SQUARE, 0, 0); // traveltime has to be set by 'link'
    setType(type);
    setPSize(size);
    ((VectorRep) getPeer()).setPaint(col);
    Payload = new ArrayList();
    }

/** This constructor is used to create packages that have an image
  as a physical representation.
  @param type The package type
  @param size The PSize of the package, <i>not</i> the physical size in
  the animation
  @param img The bitmap that will represent the package on screen
  @param sel The bitmap to show when the package is selected */
public Package(int type, int size, Image img, Image sel)
  throws PropertyVetoException
    {
    super();
    setType(type);
    setPSize(size);
    BitmapRep r = new BitmapRep(img);
    setPeer(r);
    r.setSelectedImage(sel);
    Payload = new ArrayList();
    }

/** This package constructor gives us a package that is suitable for routing
 *  with SoftwareHandlers. */
public Package(int type, String target, String origin, int size, Image img, Image sel) throws PropertyVetoException
  {
  this(type, size, img, sel);
  this.addPayload(new IDPayload(target));
  this.addPayload(new IDPayload(origin));
  }


/** This method is called by Link when a package is placed on it.
  Packages use it to react to being placed on the link. While the
  method is empty in its original form, child classes might use it
  (e.g. to increase hop counters, introduce bit-errors based on
  ErrorRate etc.) */
public void assessLink(Link l)
    {
    // basic packages ignore link
    }


////////// gets and sets ///////////////////

/** Set type of the package. This is application dependent, i.e. which
    number requires which interpretation is up to the user. This value
    is used intensively by the SimpleParseHandler class */
public void setType(int i) throws PropertyVetoException
    {
    // checkEnd();
    Type = i;
    }

/** Returns the package type */
public int getType()
    {
    return Type;
    }

/** The payload can be empty but can be used by applications to simulate
    package content. It is an array of objects. */
public void addPayload(BasicPayload data) throws PropertyVetoException
    {
    if (data == null)
      throw new PropertyVetoException("null not allowed!",
        new PropertyChangeEvent(this, "Payload", null, null));
    Payload.add(data);
    }

/** Sets the payload value at the given position to the one passed as parameter.
 *  Note that this may easily result in a nullpointer exception if there previously
 *  was no entry at that position. */
public void setPayload(int pos, BasicPayload data) throws IndexOutOfBoundsException
  {
  Payload.set(pos, data);
  }

/** Removes a payload entry from the list of payload entries */
public void removePayload(int pos) throws IndexOutOfBoundsException
  {
  Payload.remove(pos);
  }

/** Returns the package payload */
public BasicPayload getPayload(int pos) throws IndexOutOfBoundsException
  {
  return ((BasicPayload) Payload.get(pos));
  }

/** Returns all package payload in an array */
public BasicPayload[] getPayload()
  {
  int max = Payload.size();
  BasicPayload[] ret = new BasicPayload[max];
  for (int i=0; i < max; i++)
    {
    ret[i] = (BasicPayload) Payload.get(i);
    }
  return ret;
  }


/** Returns the number of lines of payload */
public int getPayloadSize()
  {
  return Payload.size();
  }

/** Set package size. This value is used by links to determine bandwidth
    usage or can be used to determine time lag in the ComplexRouter class. Note
    that if the package is already moving, this change will have no effect.
    This makes sense - the size only comes into consideration when the package
    is put onto a link.
    The obvious connection between Payload and PSize is not implemented to keep
    the system more flexible, i.e. adding additional payload lines will not
    change package size.  */
public void setPSize(int i) throws PropertyVetoException
    {
    // checkEnd();
    checkNeg(i, PSize);
    PSize = i;
    }

/** Returns package size */
public int getPSize()
    {
    return PSize;
    }

/** This method tests whether it is ok to change the values at the
  given point of time... (this method is currently not used by the class
  but may be reactived by uncommenting it in the setX()-Methods.)
  */
private void checkEnd() throws PropertyVetoException
    {
      if (Completion != 0 && Completion != TravelTime)
	throw new PropertyVetoException("Change impossible during travel!",
					null);
    }

public void move()
  {
  super.move();
  if (Completion == LossPoint) super.getList().removeElement(this);
  }

/** This method tests whether values are negative.
 @param old Former value
 @param neu The new value which is tested for negativity. */
private void checkNeg(int old, int neu) throws PropertyVetoException
    {
    if (neu < 0)
      throw
      new PropertyVetoException("Negative Values not allowed!",
	   new PropertyChangeEvent(this, "??",
				   new Integer(old), new Integer(neu)));
    }

/** When this method is called the package will be lost at a random point
 * of completion. This call cannot be undone, i.e. after calling the method
 * the package will be lost sooner or later. To be sure that the package really
 * gets lost, the method should be called before it starts to move, otherwise
 * it might be already beyond the random loss point.
 */
public void willBeLost()
  {
  LossPoint = (int)(Math.random() * (TravelTime / 3.0) + (TravelTime / 3.0));
  // ensure that package is lost somewhere in the middle.....
  }

public JComponent getCustomizer(boolean editable)
    {
    PackageCustomizer c = new PackageCustomizer(editable);
    c.setObject(this);
    return c;
    }

/** The detailed info on this package as a string */
public String toString()
    {
    String ret = "Package (" + Type + ") Size: " + PSize;
    return ret;
    }

public Object clone()
    {
    Package p = (Package) super.clone();
    p.Payload = new ArrayList();
    Iterator it = Payload.iterator();
    BasicPayload pay = null;
    while(it.hasNext())
      {
      pay = (BasicPayload) it.next();
      try
        {
        p.addPayload(pay);
        } catch (java.beans.PropertyVetoException ex)
          {
          System.err.println(this+ " has some serious programming flaw!");
          }
      }
    return p;
    }
public SimpleAnimList getAnimList()
    {
      return (SimpleAnimList)this.getList();
    }
}





