
package viror.teach.anim.net;

import viror.teach.anim.*;
import java.awt.Image;
import java.util.*;
import java.beans.PropertyChangeEvent;
import javax.swing.JComponent;
import gitzel.util.IniReader;
import java.net.URL;
import java.io.IOException;

/** This class represents a router in the animation.
  It can connect to other machines, send messages and receive
  packages from others. */
public class Router extends AbstractRouter
{

/** The IP-Address of this host, as long. This field can be used to
 have simple addresses (1, 2, 3...) for routing or can be used
 for fully fledged IP with local address etc. */
long IPAddress;

/** A short description of this router */
String Description;

/** A list of all links that leave from this machine. */
protected List LinkList;

/** A list of InternalLinks (i.e. links leading to software components
 *  rather than other routers) */
protected List ILinkList;

/** The incoming packages are delegated to this "operating system"
  running on the machine. If it is null, packages will be ignored. */
PackageHandler Handler;

/** Which customizer to use? */
private boolean Verbose = true;

/** The layer on which packages are displayed when originating from
  this router */
int PackageLayer = 1;

  //////////////////////////////////////////

/** The constructor sets up a router. It will be represented by
  the bitmap image.
  @param picture The image that represents the router
  @param select  The image that will be displayed when the
              router is selected.
  @param x The x grid coordinate
  @param y The y grid coordinate*/
public Router(Image picture, Image select, int x, int y)
    {
    super(picture, select, true, x, y);
    IPAddress = 0;
    setName("Nameless");
    Description = "No description";
    LinkList  = new ArrayList();
    ILinkList = new ArrayList();
    Handler = new SimpleParseHandler(this);
    setIdentification("Router");
    }

public Router(Image picture, Image select)
    {
    this (picture, select, 0, 0);
    }




  //////////////////////////////////////////


/** This method is called when a package arrives at the router */
public void packageArrived(Package p)
    {
    System.out.println("Router: Arrival of "+p);
    super.packageArrived(p);
    Handler.handlePackage(p);
    }

/** Calling this method puts the given package on the link. Note that
  the package will be added into the AnimList as well, so it should
  not be in there yet. The layer chosen for the package is given in
  the PackageLayer variable.
  @param p The package to be send
  @param l The number of the link on which to send the package
  @returns True if the send was successful, false otherwise. */
public boolean sendPackage(Package p, int l)
    {
    Link out = (Link) LinkList.get(l);
    if (out == null) return false;
    getList().addElement(PackageLayer, p);
    out.placePackage(this, p);
    return true;
    }

/** Calling this method sends a package via an internal link. Note that
  the package will be added into the AnimList as well, so it should
  not be in there yet. The layer chosen for the package is given in
  the PackageLayer variable.
  @param p The package to be send
  @param intLine The number of the InternalLink on which to send the package
  @returns True if the send was successful, false otherwise. */
public boolean sendInternalPackage(Package p, int intLine)
  {
  InternalLink out = (InternalLink) ILinkList.get(intLine);
  if (out == null) return false;
  getList().addElement(PackageLayer, p);
  out.placePackage(this, p);
  return true;
  }


/** Returns the number of links that leave from this machine */
public int getNumLinks() { return LinkList.size(); }

/** Returns the number of InternalLinks that connect SoftwareComponents to this
 *  machine. */
public int getNumInternalLinks() { return ILinkList.size(); }

/** Connects an element to this router. According to identification it will be placed
 *  into different lists
 */
protected void connect(AnimPath p)
  {
  if (p.getIdentification().equals("Link"))
    {
    if (!LinkList.contains(p)) LinkList.add(p);
    }
  if (p.getIdentification().equals("InternalLink"))
    {
    if (!ILinkList.contains(p)) ILinkList.add(p);
    }
  }

/** Disconnects an element from this router. */
protected void disconnect(AnimPath p)
  {
  if (p.getIdentification().equals("Link"))
    {
    LinkList.remove(p);
    }
  if (p.getIdentification().equals("InternalLink"))
    {
    ILinkList.remove(p);
    }
  }

/** Returns the router that will be reached by link number l */
public Router getGoal(int l)
    {
    Link lnk = ((Link) LinkList.get(l));
    if (lnk.getStart() == this) return (Router) lnk.getEnd();
    else return (Router) lnk.getStart();
    }

/** returns the software component that will be reached by internal link l */
public SoftwareComponent getInternalGoal(int l)
    {
    InternalLink lnk = ((InternalLink) ILinkList.get(l));
    if (lnk.getStart() == this) return (SoftwareComponent) lnk.getEnd();
    else return (SoftwareComponent) lnk.getStart();
    }

/** Returns the last link that connects to the given goal. (Most
  applications make it very unlikely to have more than one
  link to a given goal though) Only
  direct connections will be considered. If router is not in
  list -1 is returned. */
public int getLink(Router r)
    {
    Link l = null;
    Iterator it = LinkList.iterator();
    while (it.hasNext())
      {
      l = (Link) it.next();
      if (l.getStart() == this && l.getEnd() == r)
	return LinkList.lastIndexOf(l);
      if (l.getEnd() == this && l.getStart() == r)
	return LinkList.lastIndexOf(l);
      }
    return -1;
    }

/** Returns the last internal link that connects to the given goal. (Most
  applications make it very unlikely to have more than one
  link to a given goal though) Only
  direct connections will be considered. If not in
  list -1 is returned. */
public int getInternalLink(SoftwareComponent r)
    {
    InternalLink l = null;
    Iterator it = ILinkList.iterator();
    while (it.hasNext())
      {
      l = (InternalLink) it.next();
      if (l.getStart() == this && l.getEnd() == r)
	return ILinkList.lastIndexOf(l);
      if (l.getEnd() == this && l.getStart() == r)
	return ILinkList.lastIndexOf(l);
      }
    return -1;
    }

/** Sets the IP address for this router */
public void setIPAddress(long ip)
    {
    IPAddress = ip;
    }

/** Returns the IPAddress of this router */
public long getIPAddress()
    {
    return IPAddress;
    }

/** Sets the hostname for this router. This is identical to setName. */
public void setHostname(String name)
    {
    setName(name);
    }

/** Returns this host's name */
public String getHostname()
    {
    return getName();
    }

/** Reads hostname, ip address and description from an 'ini file'. The syntax of
 *  the ini file should be the following:
 *  Hostname: X
 *  IPAddress: addr
 *  Description: bla bla
 *  Description: next line of bla bla */
public void initFromURL(URL data) throws IOException, NumberFormatException
  {
  IniReader i = new IniReader(data.openStream());
  setHostname(i.getEntry("Hostname"));
  setIPAddress(new Long(i.getEntry("IPAddress")).longValue());
  setDescription(i.getEntry("Description"));
  }

/** Sets the package handler. Also the reference 'Home' is set
 to this router. */
public void setHandler(PackageHandler h)
    {
    Handler = h;
    h.setHome(this);
    }

public PackageHandler getHandler()
    {
    return Handler;
    }

public void setPackageLayer(int i)
    {
    PackageLayer = i;
    }

public int getPackageLayer()
    {
    return PackageLayer;
    }

  //////////////////////////////////////

/** Links are not cloned, only the basic properties of this router */
public Object clone()
    {
    Router r = (Router) super.clone();
    r.LinkList = new ArrayList();
    return r;
    }

public String toString()
    {
    String ret = getName() + "(" + IPAddress + ")";
    return ret;
    }

public JComponent getCustomizer(boolean editable)
    {
    UndoCustomizer r = null;
    if (Verbose) r = new RouterCustomizer(editable);
    else r = new RouterSummary();
    r.setObject(this);
    return (JComponent) r;
    }

/** This method returns true, when the large full-information Customizer is to
 *  be shown. If false a small text summary will be displayed instead. Note that
 *  the non-verbose customizer is not editable whatever the setting of edit.
 *  It is not meant for customizing but rather to have a manageable data panel
 *  for router display that does not jam half of the screen with the usually
 *  quite large routing table. */
public boolean isVerbose()
  {
  return Verbose;
  }

/** This v = true, then the large full-information Customizer is to
 *  be shown. If false a small text summary will be displayed instead. Note that
 *  the non-verbose customizer is not editable whatever the setting of edit.
 *  It is not meant for customizing but rather to have a manageable data panel
 *  for router display that does not jam half of the screen with the usually
 *  quite large routing table. */
public void setVerbose(boolean v)
  {
  Verbose = v;
  }

public void propertyChange(PropertyChangeEvent e)
    {
    super.propertyChange(e);
    // remove lost links from list
    if (!(e.getPropertyName().equals("State")) ||
	!(e.getSource() instanceof Link)) return;
    int newval = ((Integer) e.getNewValue()).intValue();
    if ((newval & AnimElement.DELETED) != 0)
      {
      LinkList.remove((Link) e.getSource());
      }
    }
}




