

package viror.teach.anim.net;

import viror.teach.anim.*;
import java.awt.*;
import java.util.*;


/**
 * This class stands for a software component that is installed on a router.
 * SoftwareComponents can communicate with a router or with each other,
 * by using sendPackage and packageArrived. In fact they are very similar to
 * routers and can be used nearly as freely - any limitations (i.e. can communicate
 * only with components on same router) are purely context specific and it is
 * the duty of the library user to avoid those.
 * This class is abstract - the two main implementations are OOComponent and
 * OrnamentalComponent. They vary in their customizers and the OOComponent also
 * has some functionality like methods to call etc...
 * @author Ralf Gitzel
 * @version 1.0
 */
public abstract class SoftwareComponent extends AbstractRouter implements Progressive, TimerListener
{

/** The name of this software component */
private String Name;

/** The set of all ProgressIndicators observing this object */
private Set PI = new HashSet();

/** All internal links this object is connected to. */
private java.util.List Links = new ArrayList();

/** The index of the link connecting to the router. -1 means that no
 * router is connected yet. (And maybe never will connect) */
private int RouterIndex = -1;

/** A reference to the router this is connected to. If no router exists, this
 *  member will be null. */
private Router Home = null;

/** The layer on which packages should be placed when they are sent. This decides
 *  whether packages will be in front of or behind routers. */
private int PackageLayer = 1;

/** A basic constructor that can be used in conjunction with StandardButtons
 *  or any other constellation where the position of the element will be
 *  set afterwards.
 */
public SoftwareComponent(Image basic, Image selected)
  {
  this(basic, selected, 0, 0);
  }

/** This constructor is the one that is most likely to be used in the case of
 *  of 'manual' use.
 */
public SoftwareComponent(Image basic, Image selected, int x, int y)
  {
  super(basic, selected, true, x, y);
  Name = "Component";
  setIdentification("SoftwareComponent");
  }

/** This method sends a package to the router in order to have it moved
 *  to other routers and then finally to the intended component.
 *  The package must conform to the routing standards of the handler used. */
public void sendPackage(Package p)
  {
  InternalLink l = getRouterLink();
  l.getList().addElement(PackageLayer, p);
  l.placePackage(this, p);
  }

/** This method sends an internal package. Parameters are as with sendPackage(). */
public void sendInternalPackage(Package p)
  {
  System.err.println("Send internally: NOT IMPLEMENTED YET!");
  }

/** Returns the router this component is connected to. If it is connected to
 *  more than one Router (which is not really a supported case) the first one
 *  will be returned. (If you have to deal with the acutal implementation of
 *  this class it might be helpful to look at the comment for getRouterIndex
 *  to better understand what is happening here.) */
public Router getHome()
  {
  if (Home == null) getRouterLink();
  return Home;
  }

/**  Returns the Link to the Router. If there is no such link, null is returned.
 *  Implementation Note: The update of RouterIndex occurs here since the links
 *  might not have a target yet when hooked up to this object. If a link is
 *  removed, RouterIndex is set to -1 again.
 *  As a side-note, the member Home is updated in this method as well - therefore
 *  this method is called in getHome() if the Home == null */
public InternalLink getRouterLink()
  {
  if (RouterIndex == -1)
    {
    // update on the fly!
    Iterator it = Links.iterator();
    InternalLink l = null;
    AnimStatic other = null;
    while(it.hasNext())
      {
      l = (InternalLink) it.next();
      if (l.getEnd()   == this) other = l.getStart();
      if (l.getStart() == this) other = l.getEnd();
      if (other instanceof Router)
        {
        RouterIndex = Links.indexOf(l);
        Home = (Router) other;
        break;
        }
      }
    }
  if (RouterIndex != -1) return (InternalLink) Links.get(RouterIndex);
  return null;
  }


public void addProgressIndicator(ProgressIndicator i)
  {
  PI.add(i);
  }

public void removeProgressIndicator(ProgressIndicator i)
  {
  PI.remove(i);
  }

public void fireProgress(int stateAffected, boolean forward)
  {
  Iterator it = PI.iterator();
  ProgressIndicator prog = null;
  while (it.hasNext())
    {
    prog = (ProgressIndicator) it.next();
    if (forward) prog.advance(stateAffected);
    else prog.stepBack(stateAffected);
    }
  }

public void fireProgress(int stateAffected, int line)
  {
  Iterator it = PI.iterator();
  ProgressIndicator prog = null;
  while (it.hasNext())
    {
    prog = (ProgressIndicator) it.next();
    prog.setPosition(stateAffected, line);
    }
  }

/** This method is called by AnimPath's setStart/setEnd(). */
protected void connect(AnimPath l)
  {
  Links.add(l);
  }

/** This method is called by Animpath's setStart/setEnd(). */
protected void disconnect(AnimPath l)
  {
  Links.remove(l);
  RouterIndex = -1;
  }

/** If the CodeValue of the timer event is 0, the package will be sent, otherwise
 *  a progress will be fired according to package type (a lookup will occur). */
public void timerEventOccured(TimerEvent t)
  {
  if (t.CodeValue == 0) sendPackage(t.Send);
  else
    {
    TableEntry e = this.progressTableLookup(t.Send.getType());
    if (e.Forward == FORWARD)
      for (int i=0; i < e.Steps; i++) this.fireProgress(e.Channel, true);
    if (e.Forward == BACK)
          for (int i=0; i < e.Steps; i++) fireProgress(e.Channel, false);
    if (e.Forward == ABSOLUTE) this.fireProgress(e.Channel, e.Steps);
    }
  }


/** remove the package that has arrived! */
public void packageArrived(Package p)
  {
  super.packageArrived(p);
  // under certain circumstances the element might be removed already
  // e.g. its arrival deleted a path (Possible with SoftwareComponents)
  if (p.getPath() != null) p.getPath().removeActor(p);
  if (p.getList() != null) p.getList().removeElement(p);
  }

//////////// set and get ////////////

public void setName(String name)
  {
  Name = name;
  }

public String getName()
  {
  return Name;
  }

public void setPackageLayer(int i)
  {
  PackageLayer = i;
  }

public int getPackageLayer()
  {
  return PackageLayer;
  }




////////////////////////////

public String toString()
  {
  String ret = "Software Component["+Name+"]";
  return ret;
  }

/** If the name of two components is equal, this method returns true. All other
 *  values are ignored. */
public boolean equals(SoftwareComponent o)
  {
  if (o.Name.equals(Name)) return true;
  return false;
  }

}