

package viror.teach.anim.net;

import javax.swing.JComponent;
import viror.teach.anim.*;
import java.util.*;
import gitzel.util.*;
import java.io.InputStream;

/** This is another PackageHandler-Implementation. It is used to pass
 *  packages to various software components based on their payload.
 *  Basically the class checks whether the i-th payload (set with IDPos)
 *  is an instance of IDPayload. If it is the event is delegated to the
 *  SoftwareComponent described there.
 *  This SoftwareHandler should be used when object oriented or component
 *  based problems are to be illustrated with the animation. Since the
 *  class inherits from PackageHandler, the normal functionality is still
 *  there.
 *  The idea is, that first the ID is checked, which will send packages to
 *  the components. If no ID is present, normal routing will be used. If
 *  a packages arrives FROM a component, the ID check will not be performed.
 *  This way components can send messages to other routers (by putting in the
 *  right package type and then sending to their router) where it will be handed
 *  to the proper component...
 *  Packages must conform to the following syntax, otherwise they will be
 *  discarded and all PackageErrorListeners that have registered with this
 *  component will be notified.
 *      PayloadID    - target
 *      PayloadID    - sender
 *
 * <p> As a side note please remember that this is a subclass of the
 * SimpleParseHandler and as such has all the normal routing functionality which
 * is only overridden if the target package is connected to this router. Also
 * note that initRoutingFromURL can be called here as well. */
public class SoftwareHandler extends SimpleParseHandler
                             implements PackagePedanticHandler
{

/** The Map that is responsible for mapping IDPayloads to SoftwareComponents. */
private Map LookUp;

/** This Map maps InternalLinks to the SoftwareComponent reachable by them... */
private Map InterLinks;

/** The set of all PackageErrorListeners */
private Set ErrorSet;

public SoftwareHandler(Router home)
  {
  super(home);
  LookUp = new HashMap();
  ErrorSet    = new HashSet();
  }

public void handlePackage(Package p)
  {
  // first check if package should cause an error an be discarded...
  if (!checkSyntax(p))
    {
    firePackageError(p, this, 0);
    return;
    }
  IDPayload sender = (IDPayload) p.getPayload(1);
  IDPayload target = (IDPayload) p.getPayload(0);
  // now check if sender/target is on this machine....
  boolean s = false, t = false;
  int out = 0;
  for (int i=0; i < getHome().getNumInternalLinks(); i++)
    {
    if (sender.isTarget(getHome().getInternalGoal(i)))
      {
      s = true;
      break;
      }
    if (target.isTarget(getHome().getInternalGoal(i)))
      {
      t = true;
      out = i;
      break;
      }
    }
  // cases:
  if (!s && !t)
    {
    // neither is on this machine, route by type via links
    // this happens if we are just an in-between node
    super.handlePackage(p);
    }
  if (s && !t)
    {
    // sender is on machine, target not.... route via links
    super.handlePackage(p);
    }
  if (t)
    {
    // target on this machine
    p.getPath().removeActor(p);
    Package o = (Package) p.clone();
    p.getList().removeElement(p);
    getHome().sendInternalPackage(o, out);
    }
  }


public String toString()
  {
  return "SoftwareHandler running on "+getHome();
  }

public JComponent getCustomizer(boolean edit)
  {
    //TODO: Implement this viror.teach.anim.net.PackageHandler method
  return null;
  }

public boolean checkSyntax(Package p)
  {
  if (!(p.getPayload(0) instanceof IDPayload)) return false;
  if (!(p.getPayload(1) instanceof IDPayload)) return false;
  return true;
  }

public void firePackageError(Package p, PackagePedanticHandler origin, int code)
  {
  Iterator it = ErrorSet.iterator();
  while (it.hasNext())
    {
    ((PackageErrorListener) it.next()).errorOccured(p, origin, code);
    }
  }

public void addErrorListener(PackageErrorListener r)
  {
  ErrorSet.add(r);
  }

public void removeErrorListener(PackageErrorListener r)
  {
  ErrorSet.remove(r);
  }


}