
package viror.teach.anim.net;

import javax.swing.*;
import java.awt.Dimension;
import viror.teach.anim.ProgressIndicator;
import java.util.*;
import java.net.*;
import gitzel.util.*;

/** <p>This class is an advanced version of the HighlightText class. The
 *  SourceCodeViewer can store several source codes and will advance
 *  between them according to a preset pattern (Script). Note that there
 *  is no actual parsing - the succession of line highlighting will always
 *  be the same and jump between different source codes are possible at any
 *  time without any regard to code (i.e. method calls etc.)
 *  Scripts consist of several table entries like this:
 *  <p><pre>              STEP    NEXTLINE(SOURCE,LINE)</pre>
 * <p>To illustrate usage of this consider the following example...
 * advance() is called and we reach step 5 of our internal position counter.
 * After looking into the Script, we find that we should display the LINE on
 * page SOURCE we find in this entry. Thus we can pretend to have a program
 * running.
 */
public class SourceCodeViewer extends JTabbedPane implements ProgressIndicator
{

  /** The number of source codes shown -1 */
  private int num = 0;

  /** The position in the source code - this is mapped to a source code page
   *  and line with the help of the ArrayList Script. */
  private int Position;

  /** The Script of this object - see intro text for details. */
  private ArrayList Script;


  public SourceCodeViewer()
  {
  Script = new ArrayList();
  //this.setPreferredSize(new Dimension(200,200));
  }

  /** add a new Source Code to the Viewer. It will be added behind the
   *  last position currently used.
   *  @param name The name of this source code - this will be displayed in
   *     the tab.
   *  @param icon The icon to be used in the tab. 'null' denotes that no icon
   *     should be used.
   *  @param text The source code lines. Use a string for each line and no
   *     \n etc. since the display is rather simple in design.
   */
  public void addSource(String name, Icon icon, String[] text)
    {
    HighlightText hlt = new HighlightText();
    hlt.setLines(text);
    add(name, hlt);
    if (icon != null) setIconAt(num, icon);
    num++;
    }

  /** Change a script entry to the given values. If entry did not exist before
   *  it will be created now.   */
  public void setEntry(int position, String source, int line)
    {
    ScriptEntry e = new ScriptEntry();
    int theOne = 0;
    for (int i=0; i<num;i++)
      if (this.getTitleAt(i).equals(source))
        {
        theOne = i;
        break;
        }
    e.Line = line;
    e.SourceCode = (HighlightText) this.getComponent(theOne);
    Script.set(position, e);
    }

  /** Returns the script entry at position i. If entry does not exist a null
   *  is returned. This is meant primarily for internal use and thus is not
   *  public. */
  protected ScriptEntry getEntry(int pos)
    {
    if (pos >= Script.size()) return null;
    return (ScriptEntry) Script.get(pos);
    }

  /** Initializes the script from an url. The script format is the following:
   *  <pre>
   *    source1 line1
   *    source2 line2
   *  <pre>
   *  Meaning; in the first step we jump to source1 in line 1, then to source2 at
   *  line2 etc. */
  public void initFromURL(URL url) throws java.io.IOException, MalformedURLException
    {
    BasicScriptReader red = new BasicScriptReader(url.openStream());
    Iterator it = red.iterator();
    String[] data = null;
    int line = 0, source = 0;
    while (it.hasNext())
      {
      data = (String[]) it.next();
      if (data.length != 2)
        {
        System.err.println(url + ": corrupt script, format:\n\tsource line\n\tsource line ...");
        return;
        }
      try
        {
        line   = new Integer(data[1]).intValue();
        source = new Integer(data[0]).intValue();
        } catch (NumberFormatException e)
          {
          System.err.println(url + ": Only integers allowed in script!");
          return;
          }
      Script.add(new ScriptEntry((HighlightText) getComponentAt(source), line));
      }
    }

  ////////////////////////////////

  /** As with HighlightText the parameter is ignored since it is only used
   *  for ProgressIndicators with multiple values.
   */
  public void advance(int i) throws ArrayIndexOutOfBoundsException
  {
  ScriptEntry e = getEntry(Position);
  if (e == null) return;
  Position++;
  e.SourceCode.setPosition(0, e.Line);
  setSelectedComponent(e.SourceCode);
  }

  public void stepBack(int i) throws ArrayIndexOutOfBoundsException
  {
  Position--;
  ScriptEntry e = getEntry(Position);
  e.SourceCode.setPosition(0, e.Line);
  setSelectedComponent(e.SourceCode);
  }

  /** Note that the first line is 0. Again the parameter i is ignored. */
  public void setPosition(int i, int pos)
  {
  Position = pos;
  ScriptEntry e = getEntry(Position);
  e.SourceCode.setPosition(0, e.Line);
  setSelectedComponent(e.SourceCode);
  }

/** The parameter i is ignored */
public int getPosition(int i)
  {
  return Position;
  }


  /////////////////////////////////

  /** Internally used class that holds lines of the Script ArrayList */
  class ScriptEntry
    {
    public HighlightText SourceCode;
    public int Line;

    public ScriptEntry(HighlightText source, int line)
      {
      SourceCode = source;
      Line = line;
      }

    public ScriptEntry()
     {
     }
    }
}